/*
 * block_io.c
 *
 * Copyright 2004 Nigel Cunningham <ncunningham@linuxmail.org>
 *
 * Distributed under GPLv2.
 * 
 * This file contains block io functions for suspend2. These are
 * used by the swapwriter and it is planned that they will also
 * be used by the NFSwriter.
 *
 */

#include <linux/suspend-common.h>
#include "block_io.h"

/* Bits in struct io_info->flags */
#define IO_WRITING 1
#define IO_RESTORE_PAGE_PROT 2
#define IO_AWAITING_READ 3
#define IO_AWAITING_WRITE 4
#define IO_CLEANUP_IN_PROGRESS 5
#define IO_HANDLE_PAGE_PROT 6

/*
 * ---------------------------------------------------------------
 *
 *     IO in progress information storage and helpers
 *
 * ---------------------------------------------------------------
 */

struct io_info {
	SYS_IO_STRUCT * sys_struct;
	SYS_ZONE_TYPE blocks[PAGE_SIZE/512];
	struct page * buffer_page;
	struct page * data_page;
	unsigned long flags;
	DEVICE_BLOCK_TYPE dev;
	int blocks_used;
	int block_size;
	struct list_head list;
	int readahead_index;
};

static LIST_HEAD(ioinfo_free);
static LIST_HEAD(ioinfo_ready_for_cleanup);
static LIST_HEAD(ioinfo_busy);
static spinlock_t ioinfo_lists_lock = SPIN_LOCK_UNLOCKED;

/* [Max] number of I/O operations pending */
static int outstanding_io = 0;
static int max_outstanding_io = 0;
static int buffer_allocs, buffer_frees;

/* [Max] number of pages used for above struct */
static int infopages = 0;
static int maxinfopages = 0;

/*
 * cleanup_one
 * 
 * Description: Clean up after completing I/O on a page.
 * Arguments:	struct io_info:	Data for I/O to be completed.
 */
static inline void cleanup_one(struct io_info * io_info)
{
	struct page * buffer_page;
	struct page * data_page;
	char *buffer_address, *data_address;
	int reading;

	buffer_page = io_info->buffer_page;
	data_page = io_info->data_page;

	/* 
	 * Already being cleaned up? Can't happen while we're single
	 * threaded, but a good check for later.
	 */
	
	if (test_and_set_bit(IO_CLEANUP_IN_PROGRESS, &io_info->flags))
		return;

	reading = test_bit(IO_AWAITING_READ, &io_info->flags);
	printlog(SUSPEND_IO, SUSPEND_HIGH, "Cleanup IO: [%p]\n", 
		io_info);

	if (reading) {
		/*
		 * Copy the page we read into the buffer our caller provided.
		 */
		data_address = (char *) kmap(data_page);
		buffer_address = (char *) kmap(buffer_page);
		memcpy(data_address, buffer_address, PAGE_SIZE);
		flush_dcache_page(data_page);
		kunmap(data_page);
		kunmap(buffer_page);
	
	}

	/* Sanity check */
	if (page_count(buffer_page) != 2)
		printk(KERN_EMERG "Cleanup IO: Page count is %d. Not good!\n",
				page_count(buffer_page));
	put_page(buffer_page);
	__free_pages(buffer_page, 0);
	buffer_frees++;
	
	outstanding_io--;
	bio_put(io_info->sys_struct);
	io_info->sys_struct = NULL;
	io_info->flags = 0;
}

/* 
 * ioinfo_cleanup_one
 *
 * Description:	Clean up a completed I/O, if one is ready.
 * Returns:	Returns pointer to the page which was cleaned up on success,
 * 		-ENODATA if none ready for cleanup.
 */

static struct io_info * ioinfo_cleanup_one(void)
{
	struct io_info * this;
	int readahead_index;
	unsigned long flags;

	/* Anything to do? */
	if (list_empty(&ioinfo_ready_for_cleanup))
		return ERR_PTR(-ENODATA);

	/* 
	 * Yes. Get our target and remove it from the list to ensure no one
	 * else duplicates our labours.
	 */
	spin_lock_irqsave(&ioinfo_lists_lock, flags);
	this = list_entry(ioinfo_ready_for_cleanup.next, struct io_info, list);
	list_del(&this->list);
	spin_unlock_irqrestore(&ioinfo_lists_lock, flags);

	/*
	 * If this I/O was a readahead, remember its index and reset it.
	 */
	readahead_index = this->readahead_index;
	this->readahead_index = -1;

	/*
	 * Do the cleanup.
	 */
	cleanup_one(this);

	/*
	 * Record the readahead as done.
	 */
	if (readahead_index > -1) {
		int index = readahead_index/(8 * sizeof(unsigned long));
		int bit = readahead_index - (index * 8 * sizeof(unsigned long));
		spin_lock_irqsave(&suspend_readahead_flags_lock, flags);
		set_bit(bit, &suspend_readahead_flags[index]);
		spin_unlock_irqrestore(&suspend_readahead_flags_lock, flags);
	}

	/*
	 * Add it to the free list.
	 */
	spin_lock_irqsave(&ioinfo_lists_lock, flags);
	list_add_tail(&this->list, &ioinfo_free);
	spin_unlock_irqrestore(&ioinfo_lists_lock, flags);
	return this;
}

/*
 * get_io_info_struct
 *
 * Description:	Get an I/O struct.
 * Returns:	Pointer to the struct prepared for use.
 */
static struct io_info * get_io_info_struct(void)
{
	unsigned long newpage = 0, flags;
	struct io_info * this = NULL;
	int iteration = 0, remaining = 0;

	do {
		if (!(iteration%8))
			RUN_IO;
		iteration++;
	
		/* We clean up an IO info struct if possible */
		ioinfo_cleanup_one();

		/* Have we reached our number-of-IOs-activate-at-one limit? */
		if ((max_async_ios) && (outstanding_io >= max_async_ios))
			continue;

		/* Can start a new I/O. Is there a free one? */
		if (!list_empty(&ioinfo_free)) {
			/* Yes. Grab it. */
			spin_lock_irqsave(&ioinfo_lists_lock, flags);
			break;
		}

		/* No. Need to allocate a new page for I/O info structs. */
		newpage = get_zeroed_page(GFP_ATOMIC);
		if (!newpage)
			continue;

		printlog(SUSPEND_MEMORY, SUSPEND_VERBOSE,
				"[NewIOPage %lx]", newpage);
		infopages++;
		if (infopages > maxinfopages)
			maxinfopages++;

		/* Prepare the new page for use. */
		this = (struct io_info *) newpage;
		remaining = PAGE_SIZE;
		spin_lock_irqsave(&ioinfo_lists_lock, flags);
		while (remaining >= (sizeof(struct io_info))) {
			list_add_tail(&this->list, &ioinfo_free);
			this = (struct io_info *) (((char *) this) + 
					sizeof(struct io_info));
			remaining -= sizeof(struct io_info);
		}
		break;
	} while (1);

	/* We have an I/O info struct. Move it to the busy list. */
	this = list_entry(ioinfo_free.next, struct io_info, list);
	list_move_tail(&this->list, &ioinfo_busy);
	spin_unlock_irqrestore(&ioinfo_lists_lock, flags);
	return this;
}

/*
 * finish_all_io
 *
 * Description:	Finishes all IO and frees all IO info struct pages.
 */
void finish_all_io(void)
{
	struct io_info * this, * next = NULL;
	unsigned long flags;

	/* Wait for all I/O to complete. */
	while (!list_empty(&ioinfo_busy))
		SYNC_IO;

	/* Clean them all up. */
	do { } while (!IS_ERR(ioinfo_cleanup_one()));

	/*
	 * We're single threaded and all I/O is completed, so we shouldn't
	 * need to use the spinlock, but let's be safe.
	 */
	spin_lock_irqsave(&ioinfo_lists_lock, flags);
	
	/* 
	 * Two stages, to avoid using freed pages.
	 *
	 * First free all io_info structs on a page except the first.
	 */
	list_for_each_entry_safe(this, next, &ioinfo_free, list) {
		if (((unsigned long) this) & ~PAGE_MASK)
			list_del(&this->list);
	}

	/* 
	 * Now we have only one reference to each page, and can safely
	 * free pages, knowing we're not going to be trying to access the
	 * same page after freeing it.
	 */
	list_for_each_entry_safe(this, next, &ioinfo_free, list) {
		list_del(&this->list);
		free_pages((unsigned long) this, 0);
		printlog(SUSPEND_MEMORY, SUSPEND_VERBOSE,
				"[FreedIOPage %lx]", this);
	}
	
	spin_unlock_irqrestore(&ioinfo_lists_lock, flags);
}

/*
 * wait_on_one_page
 *
 * Description:	Wait for a particular I/O to complete.
 */
static void wait_on_one_page(struct io_info * io_info)
{
	do {
		SYNC_IO;
	} while (ioinfo_cleanup_one() != io_info);
}

/*
 * reset_io_stats
 *
 * Description:	Reset all our sanity-checking statistics.
 */
void inline reset_io_stats(void)
{
	max_outstanding_io = outstanding_io = 0;
	maxinfopages = infopages = 0;
	buffer_allocs = buffer_frees = 0;
}

/*
 * check_io_stats
 *
 * Description:	Check that our statistics look right and print
 * 		any debugging info wanted.
 */
void suspend_check_io_stats(void)
{
	if (outstanding_io)
		printnolog(SUSPEND_IO, SUSPEND_MEDIUM, 0,
			"Outstanding_io after writing is %d.\n",
			outstanding_io);
	printnolog(SUSPEND_IO, SUSPEND_LOW, 0,
			"Maximum outstanding_io was %d.\n",
			max_outstanding_io);
	if (infopages)
		printnolog(SUSPEND_IO, SUSPEND_MEDIUM, 0,
				"Info pages is %d.\n",
				infopages);
	printnolog(SUSPEND_IO, SUSPEND_LOW, 0,
			"Max info pages was %d.\n",
			maxinfopages);
	if (buffer_allocs != buffer_frees)
		printk("Buffer allocs (%d) != buffer frees (%d)",
				buffer_allocs,
				buffer_frees);
}

/*
 * suspend_end_buffer_io_async
 *
 * Description:	Function called by block driver from interrupt context when I/O
 * 		is completed. This is the reason we use spinlocks in
 * 		manipulating the io_info lists. 		
 * 		Nearly the fs/buffer.c version, but we want to mark the page as 
 * 		done in our own structures too.
 */

static int suspend_end_bio(struct bio * bio, unsigned int num, int err)
{
	struct io_info *io_info = (struct io_info *) bio->bi_private;
	unsigned long flags;
	
	spin_lock_irqsave(&ioinfo_lists_lock, flags);
	list_move_tail(&io_info->list, &ioinfo_ready_for_cleanup);
	spin_unlock_irqrestore(&ioinfo_lists_lock, flags);
	return 0;
}

/**
 *	submit - submit BIO request.
 *	@rw:	READ or WRITE.
 *	@io_info: IO info structure.
 *
 *	Straight from the textbook - allocate and initialize the bio.
 *	If we're writing, make sure the page is marked as dirty.
 *	Then submit it and carry on.
 */

static int submit(int rw, struct io_info * io_info)
{
	int error = 0;
	struct bio * bio = NULL;

	while (!bio) {
		bio = bio_alloc(GFP_ATOMIC,1);
		if (!bio) {
			io_schedule();
			ioinfo_cleanup_one();
		}
	}

	bio->bi_sector = io_info->blocks[0] << (PAGE_SHIFT - 9);
	bio->bi_bdev = io_info->dev;
	bio->bi_private = io_info;
	bio->bi_end_io = suspend_end_bio;
	io_info->sys_struct = bio;

	if (bio_add_page(bio, io_info->buffer_page, PAGE_SIZE, 0) < PAGE_SIZE) {
		printk("ERROR: adding page to bio at %ld\n",
				io_info->blocks[0]);
		bio_put(bio);
		return -EFAULT;
	}

	if (rw == WRITE)
		bio_set_pages_dirty(bio);
	submit_bio(rw,bio);
	return error;
}

/*
 * suspend_set_block_size
 *
 * Description: Set the blocksize for a bdev. This is a separate function
 * 		because we have different versions for 2.4 and 2.6.
 */
inline int suspend_set_block_size(struct block_device * bdev, int size)
{
	return set_blocksize(bdev, size);
}

inline int suspend_get_block_size(struct block_device * bdev)
{
	return SWAP_BLOCKSIZE(bdev);
}

/*
 * start_one
 *
 * Description:	Prepare and start a read or write operation.
 * 		Note that we use our own buffer for reading or writing.
 * 		This simplifies doing readahead and asynchronous writing.
 * 		We can begin a read without knowing the location into which
 * 		the data will eventually be placed, and the buffer passed
 * 		for a write can be reused immediately (essential for the
 * 		plugins system).
 * 		Failure? What's that?
 * Returns:	The io_info struct created.
 */
static struct io_info * start_one(int rw, struct submit_params * submit_info)
{
	struct io_info * io_info = get_io_info_struct();
	unsigned long buffer_virt;
	char * to, * from;
	struct page * buffer_page;
	int i;


	/* Get our local buffer */
	printlog(SUSPEND_IO, SUSPEND_HIGH, "Start_IO: [%p]", io_info);
	while (!(buffer_virt = get_zeroed_page(GFP_ATOMIC))) {
		RUN_IO;
		ioinfo_cleanup_one();
	}
	buffer_allocs++;
	printlog(SUSPEND_IO, SUSPEND_HIGH, "[ALLOC BUFFER]->%d",
			nr_free_pages());
	buffer_page = virt_to_page(buffer_virt);
	
	/* Copy settings to the io_info struct */
	io_info->data_page = submit_info->page;
	io_info->buffer_page = buffer_page;
	io_info->readahead_index = submit_info->readahead_index;

	/* If writing, copy our data. The data is probably in
	 * lowmem, but we cannot be certain. If there is no
	 * compression/encryption, we might be passed the
	 * actual source page's address. */
	if (rw == WRITE) {
		set_bit(IO_WRITING, &io_info->flags);

		to = (char *) buffer_virt;
		from = KMAP_ATOMIC(io_info->data_page);
		memcpy(to, from, PAGE_SIZE);
		flush_dcache_page(io_info->data_page);
		flush_dcache_page(buffer_page);
		KUNMAP_ATOMIC(io_info->data_page);
	}

	/* Submit the page */
	get_page(buffer_page);
	
	io_info->dev = submit_info->dev;
	for (i = 0; i < submit_info->blocks_used; i++)
		io_info->blocks[i] = submit_info->blocks[i];
	io_info->blocks_used = submit_info->blocks_used;
	io_info->block_size = PAGE_SIZE / submit_info->blocks_used;

	if (rw == READ)
		set_bit(IO_AWAITING_READ, &io_info->flags);
	else
		set_bit(IO_AWAITING_WRITE, &io_info->flags);

	printlog(SUSPEND_IO, SUSPEND_HIGH, "-> (PRE BRW) %d\n",
			nr_free_pages());

 	submit(rw, io_info);
	
	outstanding_io++;
	if (outstanding_io > max_outstanding_io)
		max_outstanding_io++;
	
	return io_info;
}

void cleanup_completed_io(void)
{
	/*
	 * We clean up all we can at once so we don't only batch the first lot.
	 */
	do { } while (!ioinfo_cleanup_one());
}

void do_suspend_io(int rw, 
		struct submit_params * submit_info, int syncio)
{
	struct io_info * io_info = start_one(rw, submit_info);
	if (!io_info) {
		/* Wait on pages until no more */
		wait_on_one_page(ERR_PTR(-ENODATA));
	} else if (syncio) {
		//finish_all_io();
		wait_on_one_page(io_info);
	}
} 

/* We used to use bread here, but it doesn't correctly handle
 * blocksize != PAGE_SIZE. Now we create a submit_info to get the data we
 * want and use our normal routines (synchronously).
 */

void bdev_page_io(int rw, DEVICE_BLOCK_TYPE bdev, long pos,
		struct page * page)
{
	struct submit_params submit_info;

	submit_info.page = page;
	submit_info.dev = bdev;

	submit_info.blocks[0] = pos;
	submit_info.blocks_used = 1;
	submit_info.readahead_index = -1;
	do_suspend_io(rw, &submit_info, 1);
}

void suspend_wait_on_readahead(int readahead_index)
{
	int index = readahead_index/(8 * sizeof(unsigned long));
	int bit = readahead_index - index * 8 * sizeof(unsigned long);

	/* read_ahead_index is the one we want to return */
	while (!test_bit(bit, &suspend_readahead_flags[index])) {
		RUN_IO;
		ioinfo_cleanup_one();
	}

}
