/*
 * kernel/power/memory_pool.c
 *  
 * Copyright (C) 2003,2004 Nigel Cunningham <ncunningham@users.sourceforge.net>
 *
 * This file is released under the GPLv2.
 *
 * It contains routines for managing the memory pool during software suspend
 * operation.
 * 
 * The memory pool is a pool of pages from which page allocations
 * are satisfied while we are suspending, and into which freed pages are
 * released. In this way, we can keep the image size static and consistent
 * while still using normal I/O routines to save the image and while saving
 * the image in two parts.
 * 
 * During suspend, almost all of the page allocations are order zero. Provision
 * is made for one order one and one order two allocation. This provision is 
 * utilised by the swapwriter for allocating memory which is used for structures
 * containing header page. (It could be made to use order zero allocations; this
 * just hasn't been done yet).
 */

#define SWSUSP_MEMORY_POOL_C

#include <linux/suspend-common.h>

/* The maximum order we handle */
#define MAX_POOL_ORDER 2

static unsigned long * suspend_memory_pool[MAX_POOL_ORDER + 1] = 
	{ NULL, NULL, NULL };
static int suspend_pool_level_limit[MAX_POOL_ORDER + 1] = { 0, 1, 1 };
static int suspend_pool_level[MAX_POOL_ORDER + 1] = { 0, 0, 0 };
static spinlock_t suspend_memory_pool_lock = SPIN_LOCK_UNLOCKED;

/* suspend_memory_pool_level()
 *
 * Description:	Returns the number of pages currently in the pool.
 * Returns:	Int.		Number of pages in the pool.
 */
int suspend_memory_pool_level(void)
{
	int order, sum = 0;

	for (order = 0; order <= MAX_POOL_ORDER; order++)
		sum+= suspend_pool_level[order] * (1 << order);

	return sum;
}

/* display_memory_pool_pages()
 *
 * Description:	Display the current contents of the memory pool.
 */	
void display_memory_pool_pages(void)
{
	unsigned long * next, flags;
	int order, count;

	spin_lock_irqsave(&suspend_memory_pool_lock, flags);
	printlog(SUSPEND_MEM_POOL, SUSPEND_VERBOSE, "Memory pool:\n");
	for (order = 0; order <= MAX_POOL_ORDER; order++) {
		printlog(SUSPEND_MEM_POOL, SUSPEND_VERBOSE,
				"- Order %d:\n", order);
		next = suspend_memory_pool[order];
		count = 0;
		while (next) {
			printlog(SUSPEND_MEM_POOL, SUSPEND_VERBOSE,
					"[%p] ", next);
			count++;
			next = (unsigned long *) *next;
		}
		if (count)
			printlog(SUSPEND_MEM_POOL, SUSPEND_VERBOSE,
					"(%d entries)\n", count);
		else
			printlog(SUSPEND_MEM_POOL, SUSPEND_VERBOSE,
					"(empty)\n");
	}
	spin_unlock_irqrestore(&suspend_memory_pool_lock, flags);
}

/* fill_suspend_memory_pool()
 *
 * Description:	Fill the memory pool from the main free memory pool in the
 * 		first instance, or grabbed pages if that fails.
 * 		We allocate @sizesought order 0 pages, plus 1 each
 * 		of the higher order allocations.
 * Arguments:	int.		Number of order zero pages requested.
 * Returns:	int.		Number of order zero pages obtained.
 */
int fill_suspend_memory_pool(int sizesought)
{
	int i = 0, j, order, orig_state = 
		software_suspend_state & SOFTWARE_SUSPEND_USE_MEMORY_POOL;
	unsigned long *this = NULL;
	unsigned long flags;

	spin_lock_irqsave(&suspend_memory_pool_lock, flags);
	
	/* Pool must not be active for this to work */
	software_suspend_state &= ~SOFTWARE_SUSPEND_USE_MEMORY_POOL;

	suspend_pool_level_limit[0] = sizesought;

	for (order = MAX_POOL_ORDER; order >= 0; order--) {
		for (i = suspend_pool_level[order]; 
				i < suspend_pool_level_limit[order]; i++) {
			this = (unsigned long *) get_grabbed_pages(order);
			if (!this) {
				printk(name_suspend "Error. %d pages wanted of"
					" order %d for suspend memory pool,"
					" got %d.\n",
					suspend_pool_level_limit[order],
					order, i - 1);
				break;
			}
			for (j = 0; j < (1 << order); j++)
				SetPageChecksumIgnore(virt_to_page(this + j));
			*this = (unsigned long) suspend_memory_pool[order];
			suspend_memory_pool[order] = this;
			suspend_pool_level[order]++;
		}
	}

	software_suspend_state |= orig_state;

	spin_unlock_irqrestore(&suspend_memory_pool_lock, flags);

	display_memory_pool_pages();

	return 0;
}

/* empty_suspend_memory_pool()
 *
 * Description:	Drain our memory pool.
 */
void empty_suspend_memory_pool(void)
{
	unsigned long flags;
	int order;

	display_memory_pool_pages();
	spin_lock_irqsave(&suspend_memory_pool_lock, flags);

	/* Pool must not be active for this to work */
	software_suspend_state &= ~SOFTWARE_SUSPEND_USE_MEMORY_POOL;

	for (order = 0; order <= MAX_POOL_ORDER; order++) {
		while (suspend_memory_pool[order]) {
			unsigned long next = *suspend_memory_pool[order];
			if (page_count(virt_to_page((unsigned long)
					suspend_memory_pool[order])) != 1)
				printk("Error freeing page %p from memory"
					" pool. Page count is %d (should be"
					" 1).\n",
				suspend_memory_pool[order],
				page_count(virt_to_page((unsigned long)
						suspend_memory_pool[order])));
			*suspend_memory_pool[order] = 0;
			free_pages((unsigned long) suspend_memory_pool[order],
				order);
			suspend_memory_pool[order] = (unsigned long *) next;
			suspend_pool_level[order]--;
		}
	}
	
	spin_unlock_irqrestore(&suspend_memory_pool_lock, flags);
}

/* get_suspend_pool_page()
 * 
 * Description:	Our equivalent to __alloc_pages (minus zone mask).
 * 		May be called from interrupt context.
 * Arguments:	unsigned int:	Mask. We really only care about __GFP_WAIT.
 * 				We're giving normal zone pages regardless.
 * 		order:		The number of pages (1 << order) wanted.
 * Returns:	struct page *:	Pointer (possibly NULL) to pages allocated.
 */
struct page * get_suspend_pool_page(unsigned int gfp_mask, unsigned int order)
{
	unsigned long * this = NULL, flags;
	struct page * page;
	static int printederror = 0;
	static unsigned long * orig_value;
	int i;

	if (order > 0) {
		if (order <= MAX_POOL_ORDER) {
			spin_lock_irqsave(&suspend_memory_pool_lock, flags);
			if (!suspend_pool_level[order]) {
				printk("No order %d allocation available.\n",
						order);
				spin_unlock_irqrestore(
						&suspend_memory_pool_lock,
						flags);
				display_memory_pool_pages();
				return NULL;
			}
			goto check_and_return;
		}
		
		if (!printederror) {
			printederror = 1;
			printk("Swsusp memory pool:"
				" Rejecting allocation of order %d.\n", order);
		}
		display_memory_pool_pages();
		return NULL;
	} else
		orig_value = suspend_memory_pool[order];

try_again:
	if ((!suspend_memory_pool[order]) && (!(gfp_mask & __GFP_WAIT))) {
		display_memory_pool_pages();
		return NULL;
	}

	while(!suspend_memory_pool[order]) {
		active_writer->ops.writer.wait_on_io(0);
		schedule();
	}

	spin_lock_irqsave(&suspend_memory_pool_lock, flags);
	if (!suspend_memory_pool[order]) {
		spin_unlock_irqrestore(&suspend_memory_pool_lock, flags);
		goto try_again;
	}
check_and_return:
	this = suspend_memory_pool[order];
	suspend_memory_pool[order] = (unsigned long *) *this;
	*this = 0;
	suspend_pool_level[order]--;
	spin_unlock_irqrestore(&suspend_memory_pool_lock, flags);

	page = virt_to_page(this);
	for (i = 0; i < (1 << order); i++) {
		ClearPageChecksumIgnore(page + i);
		clear_page(this + i * (PAGE_SIZE / sizeof(unsigned long)));
	}
	
	if (page_count(page) != 1)
		printk("Error getting page %p from memory pool. "
			"Page count is %d (should be 1).\n",
			this,
			page_count(page));
	if (PageLRU(page))
		BUG();
	if (PageActive(page))
		BUG();
	if (suspend_pool_level[order] && !suspend_memory_pool[order]) {
		printk("Original value was %p.\n", orig_value);
		BUG();
	}

	display_memory_pool_pages();

	check_shift_keys(0, NULL);

	return page;
}

/* free_suspend_pool_pages()
 *
 * Description:	Our equivalent to  __free_pages. Put freed pages into the pool.
 * 		HighMem pages do still get freed to the normal pool because they
 * 		aren't going to affect the consistency of our image - worse case,
 * 		we write a few free pages.
 * Arguments:	Struct page *:	First page to be freed.
 * 		Unsigned int:	Size of allocation being freed.
 */
void free_suspend_pool_pages(struct page *page, unsigned int order)
{
	unsigned long *new_head, flags;
	int i;
	static unsigned long * orig_value;

	printlog(SUSPEND_MEM_POOL, SUSPEND_VERBOSE, "Freeing page %p (%p), order %d.\n",
		page_address(page), page, order);
	
	for (i = 0; i < (1 << order); i++) {
		clear_page(page_address(page + i));
		set_page_count(page + i, 0);
		SetPageChecksumIgnore(page + i);
	}

	spin_lock_irqsave(&suspend_memory_pool_lock, flags);
	
	if (order && (order <= MAX_POOL_ORDER) 
			&& (suspend_pool_level[order] < suspend_pool_level_limit[order])) {
		new_head = (unsigned long *) page_address(page);
		set_page_count(page, 1);
		*new_head = (unsigned long) suspend_memory_pool[order];
		suspend_memory_pool[order] = new_head;
		suspend_pool_level[order]++;
		spin_unlock_irqrestore(&suspend_memory_pool_lock, flags);
		display_memory_pool_pages();
		check_shift_keys(0, NULL);
		return;
	}

	if (order <= MAX_POOL_ORDER)
		orig_value = suspend_memory_pool[order];

	for (i = 0; i < (1 << order); i++) {
		new_head = (unsigned long *) page_address(page + i);
		set_page_count(page + i, 1);
		*new_head = (unsigned long) suspend_memory_pool[0];
		suspend_memory_pool[0] = new_head;
		suspend_pool_level[0]++;
	}
	spin_unlock_irqrestore(&suspend_memory_pool_lock, flags);
	
	display_memory_pool_pages();

	check_shift_keys(0, NULL);
}

int suspend_page_in_pool(struct page * page)
{
	unsigned long flags, *next, *address;
	int order;

	if (PageHighMem(page))
		return 0;

	address = (unsigned long *) page_address(page);

	spin_lock_irqsave(&suspend_memory_pool_lock, flags);
	for (order = 0; order <= MAX_POOL_ORDER; order++) {
		next = suspend_memory_pool[order];
		while (next) {
			if (next == address) {
				spin_unlock_irqrestore(&suspend_memory_pool_lock, flags);
				return 1;
			}
			next = (unsigned long *) *next;
		}
	}
	spin_unlock_irqrestore(&suspend_memory_pool_lock, flags);
	return 0;
}
