/*
 * kernel/power/pagedir.c
 *
 * Copyright (C) 1998-2001 Gabor Kuti <seasons@fornax.hu>
 * Copyright (C) 1998,2001,2002 Pavel Machek <pavel@suse.cz>
 * Copyright (C) 2002-2003 Florent Chabaud <fchabaud@free.fr>
 * Copyright (C) 2002-2004 Nigel Cunningham <ncunningham@users.sourceforge.net>
 *
 * This file is released under the GPLv2.
 *
 * Routines for handling pagesets.
 * Note that pbes aren't actually stored as such. They're stored as
 * ranges (extents is the term, I'm told).
 */

#define SWSUSP_PAGEDIR_C
#include <linux/suspend-common.h>
extern struct pagedir pagedir1, pagedir2, pagedir_resume;

/* setup_pbe_variable
 *
 * Description:	Set up one variable in a page backup entry from the range list.
 * Arguments:	unsigned long:		The variable which will contain the 
 * 					value.
 * 		struct range**:		Address of the pointer to the current
 * 					range.
 * 		struct rangechain*:	Address of the rangechain we are 
 * 					traversing.
 */
void setup_pbe_variable(unsigned long * variable, struct range ** currentrange,
		struct rangechain * chain)
{
	*currentrange = chain->first;
	if (chain->first)
		*variable = chain->first->minimum;
	else {
		*variable = 0;
		printnolog(SUSPEND_RANGES, SUSPEND_VERBOSE, 1, 
			"Initialising variable from empty chain (%s).\n",
			chain->name);
	}
}

/* get_first_pbe
 *
 * Description:	Get the first page backup entry for a pagedir.
 * Arguments:	struct pbe *:	Address of the page backup entry we're 
 * 				populating.
 * 		struct pagedir: Pagedir providing the data.
 */
void get_first_pbe(struct pbe * pbe, struct pagedir * pagedir)
{
	unsigned long currentorig, currentaddress;

	pbe->pagedir = pagedir;

	/* Get raw initial values */
	setup_pbe_variable((unsigned long *) &pbe->origaddress, 
			&pbe->currentorigrange, &pagedir->origranges);
	setup_pbe_variable((unsigned long *) &pbe->address,
			&pbe->currentdestrange, &pagedir->destranges);

	/* Convert to range values */
	currentorig = (unsigned long) pbe->origaddress;
	currentaddress = (unsigned long) pbe->address;

	pbe->origaddress = mem_map + currentorig;
	pbe->address = mem_map + currentaddress;

	if ((currentaddress < 0) || (currentaddress > max_mapnr))
		panic("Argh! Destination range value %ld is invalid!",
				currentaddress);
}

/* get_next_pbe
 *
 * Description:	Get the next page backup entry in a pagedir.
 * Arguments:	struct pbe *:	Address of the pbe we're updating.
 */
void get_next_pbe(struct pbe * pbe)
{
	unsigned long currentorig, currentaddress;

	/* Convert to range values */
	currentorig = (pbe->origaddress - mem_map);
	currentaddress = (pbe->address - mem_map);

	/* Update values */
	GET_RANGE_NEXT(pbe->currentorigrange, currentorig);
	GET_RANGE_NEXT(pbe->currentdestrange, currentaddress);
	
	pbe->origaddress = mem_map + currentorig;
	pbe->address = mem_map + currentaddress;
}

/*
 * --------------------------------------------------------------------------------------
 *
 * 	Local Page Flags routines.
 *
 * 	Rather than using the rare and precious flags in struct page, we allocate
 * 	our own bitmaps dynamically.
 * 
 */

/* ------------------------------------------------------------------------- */

/* copy_pageset1
 *
 * Description:	Make the atomic copy of pageset1. We can't use copy_page (as we
 * 		once did) because we can't be sure what side effects it has. On
 * 		my Duron, with 3DNOW, kernel_fpu_begin increments preempt
 * 		count, making our preempt count at resume time 4 instead of 3.
 */
extern char __nosavedata swsusp_pg_dir[PAGE_SIZE]
                  __attribute__ ((aligned (PAGE_SIZE)));

void copy_pageset1(void)
{
	int i = 0;
	struct pbe pbe;

	/* FIXME: Arch dependant function? */
#ifdef CONFIG_X86
	__asm__( "movl %%ecx,%%cr3\n" ::"c"(__pa(swsusp_pg_dir)));
#endif

	get_first_pbe(&pbe, &pagedir1);

	for (i = 0; i < pageset1_size; i++) {
		int loop;
		unsigned long * origpage = KMAP_ATOMIC(pbe.origaddress);
		unsigned long * copypage = page_address(pbe.address);
		
		for (loop=0; loop < (PAGE_SIZE / sizeof(unsigned long)); loop++)
			*(copypage + loop) = *(origpage + loop);
		KUNMAP_ATOMIC(pbe.origaddress);
		get_next_pbe(&pbe);
	}
	
}

/* free_pagedir
 *
 * Description:	Free a previously allocated pagedir.
 * Arguments:	struct pagedir *:	Pointer to the pagedir being freed.
 */
void free_pagedir(struct pagedir * p)
{
	PRINTFREEMEM("at start of free_pagedir");

	if (p->allocdranges.first) {
		/* Free allocated pages */
		struct range * rangepointer;
		unsigned long pagenumber;
		range_for_each(&p->allocdranges, rangepointer, pagenumber) {
			ClearPageNosave(mem_map+pagenumber);
			free_page((unsigned long) page_address(mem_map+pagenumber));
		}
	}

	suspend_store_free_mem(SUSPEND_FREE_EXTRA_PD1, 1);

	/* For pagedir 2, destranges == origranges */
	if (p->pagedir_num == 2)
		p->destranges.first = NULL;
	
	put_range_chain(&p->origranges);
	put_range_chain(&p->destranges);
	put_range_chain(&p->allocdranges);

	PRINTFREEMEM("at end of free_pagedir");
	printlog(SUSPEND_SWAP, SUSPEND_MEDIUM, "Pageset size was %d.\n", p->pageset_size);
	p->pageset_size = 0;
}

/* allocate_extra_pagedir_memory
 *
 * Description:	Allocate memory for making the atomic copy of pagedir1 in the
 * 		case where it is bigger than pagedir2.
 * Arguments:	struct pagedir *: 	The pagedir for which we should 
 * 					allocate memory.
 * 		int:			Size of pageset 1.
 * 		int:			Size of pageset 2.
 * Result:	int. Zero on success. One if unable to allocate enough memory.
 */
int allocate_extra_pagedir_memory(struct pagedir * p, int pageset_size,
		int alloc_from)
{
	int num_to_alloc = pageset_size - alloc_from - p->allocdranges.size;
	int j, order;

	prepare_status(0, 0, "Preparing page directory.");

	PRINTFREEMEM("at start of allocate_extra_pagedir_memory");

	if (num_to_alloc < 1)
		num_to_alloc = 0;

	if (num_to_alloc) {
		int num_added = 0, numnosaveallocated=0;
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
		int origallocd = p->allocdranges.size;
#endif
	
		PRINTFREEMEM("prior to attempt");

		order = generic_fls(num_to_alloc);
		if (order >= MAX_ORDER)
			order = MAX_ORDER - 1;

		while (num_added < num_to_alloc) {
			struct page * newpage;
			unsigned long virt;
			
			while ((1 << order) > (num_to_alloc - num_added))
				order--;

			virt = get_grabbed_pages(order);
			while ((!virt) && (order > 0)) {
				order--;
				virt = get_grabbed_pages(order);
			}

			if (!virt) {
				p->pageset_size += num_added;
				printlog(SUSPEND_PAGESETS, SUSPEND_VERBOSE,
					"   Allocated (extra) memory for pages"
					" from %d-%d (%d pages).\n",
					alloc_from, pageset_size - 1, 
					p->allocdranges.size - origallocd);
				printk("Couldn't get enough yet."
						" %d pages short.\n",
						num_to_alloc - num_added);
				PRINTFREEMEM("at abort of "
					"allocate_extra_pagedir_memory");
				suspend_store_free_mem(SUSPEND_FREE_EXTRA_PD1, 0);
				return 1;
			}

			newpage = virt_to_page(virt);
			for (j = 0; j < (1 << order); j++) {
				SetPageNosave(newpage + j);
				/* Pages will be freed one at a time. */
				set_page_count(newpage + j, 1);
				add_to_range_chain(&p->allocdranges, newpage - mem_map + j);
				numnosaveallocated++;
			}
			num_added+= (1 << order);
		}
		printlog(SUSPEND_PAGESETS, SUSPEND_VERBOSE,
			"   Allocated (extra) memory for pages "
			"from %d-%d (%d pages).\n",
			alloc_from, pageset_size - 1,
			p->allocdranges.size - origallocd);
	}

	p->pageset_size = pageset_size;

	suspend_store_free_mem(SUSPEND_FREE_EXTRA_PD1, 0);
	PRINTFREEMEM("at end of allocate_extra_pagedir_memory");
	return 0;
}

/* mark_page_pageset2
 * 
 * Description:	This looks a bit pointless, but it seems to be necessary for
 * 		2.6.
 * 		I'll look at it some more.
 * Arguments:	struct page *:	The page to be marked as Pageset2.
 */
static inline void mark_page_pageset2(struct page * page)
{
	SetPagePageset2(page);
}

/* mark_pages_for_pageset2
 *
 * Description:	Mark unshared pages in processes not needed for suspend as
 * 		being able to be written out in a separate pagedir.
 * 		HighMem pages are simply marked as pageset2. They won't be
 * 		needed during suspend.
 */

void mark_pages_for_pageset2(void)
{
	int i, numpageset2 = 0;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
	struct zone * zone;
#endif

	if (max_mapnr != num_physpages) {
		abort_suspend("mapnr is not expected");
		return;
	}
	
	/* Start by marking no pages as pageset2 */
	for (i = 0; i < max_mapnr; i++)
		ClearPagePageset2(mem_map+i);

	/* Add LRU pages */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,99)
	for_each_zone(zone) {
		spin_lock(&zone->lru_lock);
		INACTIVE_LIST_FOR_EACH(SetPagePageset2);
		ACTIVE_LIST_FOR_EACH(SetPagePageset2);
		spin_unlock(&zone->lru_lock);
	}
#else
	spin_lock(&pagemap_lru_lock);
	INACTIVE_LIST_FOR_EACH(SetPagePageset2);
	ACTIVE_LIST_FOR_EACH(SetPagePageset2);
	spin_unlock(&pagemap_lru_lock);
#endif

	/* Ensure range pages are not Pageset2 */
	if (num_range_pages) {
		unsigned long * range_pages;
		range_pages = get_rangepages_list();
		if (!range_pages)
			return;

		for (i = 1; i <= num_range_pages; i++) {
			struct page * page;
			page = virt_to_page(range_pages[i]);
			// Must be assigned by the time recalc stats is called
			if (PagePageset2(page)) {
				printlog(SUSPEND_PAGESETS, SUSPEND_ERROR,
					"Pagedir[%d] was marked as pageset2 -"
					" unmarking.\n", i);
				ClearPagePageset2(page);
				numpageset2--;
			}
		}
		put_rangepages_list(range_pages);
	}

	/* Finally, ensure that Slab pages are not Pageset2. */

	for (i = 0; i < max_mapnr; i++) {
		if (PageSlab(mem_map+i)) {
			if (TestAndClearPagePageset2(mem_map+i)) {
				printk("Found page %d is slab page "
						"but marked pageset 2.\n", i);
				numpageset2--;
			}
		}
#if 0
		if (PageHighMem(mem_map+i))
			if (!TestAndSetPagePageset2(mem_map+i))
				numpageset2++;
#endif
	}
}

/* warmup_collision_cache
 *
 * Description:	Mark the pages which are used by the original kernel.
 */
void warmup_collision_cache(void) {
	int i;
	struct range * rangepointer = NULL;
	unsigned long pagenumber;
	
	/* Allocatemap doesn't get deallocated because it's forgotten when we
	 * copy PageDir1 back. It doesn't matter if it collides because it is
	 * not used during the copy back itself.
	 */
	allocatemap(&in_use_map, 0);
	printlog(SUSPEND_IO, SUSPEND_VERBOSE, "Setting up pagedir cache...");
	for (i = 0; i < max_mapnr; i++)
		ClearPageInUse(mem_map+i);

	range_for_each(&pagedir_resume.origranges, rangepointer, pagenumber)
		SetPageInUse(mem_map+pagenumber);
}

/* get_pageset1_load_addresses
 * 
 * Description: We check here that pagedir & pages it points to won't collide
 * 		with pages where we're going to restore from the loaded pages
 * 		later.
 * Returns:	Zero on success, one if couldn't find enough pages (shouldn't
 * 		happen).
 */

int get_pageset1_load_addresses(void)
{
	int i, nrdone = 0, result = 0;
	void **eaten_memory = NULL, **this;
	struct page * pageaddr = NULL;

	/*
	 * Because we're trying to make this work when we're saving as much
	 * memory as possible we need to remember the pages we reject here
	 * and then free them when we're done.
	 */
	
	for(i=0; i < pagedir_resume.pageset_size; i++) {
		while ((this = (void *) get_zeroed_page(GFP_ATOMIC))) {
			memset(this, 0, PAGE_SIZE);
			pageaddr = virt_to_page(this);
			if (!PageInUse(pageaddr)) {
				break;
			}
			*this = eaten_memory;
			eaten_memory = this;
		}
		if (!this) {
			abort_suspend("Error: Ran out of memory seeking locations for reloading data.");
			result = 1;
			break;
		}
		add_to_range_chain(&pagedir_resume.destranges, pageaddr - mem_map);
		nrdone++;
	}

	/* Free unwanted memory */
	while(eaten_memory) {
		this = eaten_memory;
		eaten_memory = *eaten_memory;
		free_page((unsigned long) this);
	}
	
	return result;
}

/* set_chain_names
 *
 * Description:	Set the chain names for a pagedir. (For debugging).
 * Arguments:	struct pagedir: The pagedir on which we want to set the names.
 */

void set_chain_names(struct pagedir * p)
{
	p->origranges.name = "original addresses";
	p->destranges.name =  "destination addresses";
	p->allocdranges.name = "allocated addresses";
}
