/*
 * /kernel/power/proc.c
 *
 * Copyright (C) 2002-2003 Nigel Cunningham <ncunningham@users.sourceforge.net>
 *
 * This file is released under the GPLv2.
 *
 * This file is to realize proc entries for tuning Software Suspend.
 *
 * Versions:
 * 1: /proc/sys/kernel/suspend the only tuning interface
 * 2: Initial version of this file
 * 3: Removed kernel debugger parameter.
 *    Added checkpage parameter (for checking checksum of a page over time).
 * 4: Added entry for maximum granularity in splash screen progress bar.
 *    (Progress bar is slow, but the right setting will vary with disk &
 *    processor speed and the user's tastes).
 * 5: Added enable_escape to control ability to cancel aborting by pressing
 *    ESC key.
 * 6: Removed checksumming and checkpage parameter. Made all debugging proc
 *    entries dependant upon debugging being compiled in.
 *    Meaning of some flags also changed in this version.
 * 7: Added header_locations entry to simplify getting the resume= parameter for
 *    swapfiles easy and swapfile entry for automatically doing swapon/off from
 *    swapfiles as well as partitions.
 * 8: Added option for marking process pages as pageset 2 (processes_pageset2).
 * 9: Added option for keep image mode.
 *    Enumeration patch from Michael Frank applied.
 * 10: Various corrections to when options are disabled/enabled;
 *     Added option for specifying expected compression.
 * 11: Added option for freezer testing. Debug only.
 * 12: Removed test entries no_async_[read|write], processes_pageset2 and
 *     NoPageset2.
 * 13: Make default_console_level available when debugging disabled, but limited
 *     to 0 or 1.
 * 14: Rewrite to allow for dynamic registration of proc entries and smooth the
 *     transition to kobjects in 2.6.
 * 15: Add setting resume2 parameter without rebooting (still need to run lilo
 *     though!).
 * 16: Add support for generic string handling and switch resume2 to use it. Add
 *     pre_suspend_script and post_suspend_script parameters and async activate
 *     flag.
 */

#define SWSUSP_PROC_C

static int suspend_proc_version = 15;
static int proc_initialised = 0;

#include <linux/suspend-common.h>
#include <linux/suspend-debug.h>
#include <linux/proc_fs.h>

static struct list_head suspend_proc_entries;
static struct proc_dir_entry *suspend_dir, *compat_entry, *compat_parent;

extern void suspend_console_proc_init(void);
extern char resume_file[256];	/* For resume= kernel option */

static char suspend_core_version[] = SWSUSP_CORE_VERSION;

/*
 * suspend_write_compat_proc.
 *
 * This entry allows all of the settings to be set at once. 
 * It was originally for compatibility with pre- /proc/suspend
 * versions, but has been retained because it makes saving and
 * restoring the configuration simpler.
 */
static int suspend_write_compat_proc(struct file *file, const char * buffer,
		unsigned long count, void * data)
{
	char * buf = (char *) get_zeroed_page(GFP_ATOMIC), *lastbuf;
	int i;
	unsigned long nextval;

	if (!buf)
		return -ENOMEM;

	if (count > PAGE_SIZE)
		count = PAGE_SIZE;

	if (copy_from_user(buf, buffer, count))
		return -EFAULT;
	for (i = 0; i < 6; i++) {
		if (!buf)
			break;
		lastbuf = buf;
		nextval = simple_strtoul(buf, &buf, 0);
		if (buf == lastbuf)
			break;
		switch (i) {
			case 0:
				suspend_result = nextval;
				break;
			case 1:
				suspend_action = nextval;
				break;
			case 2:
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
				suspend_debug_state = nextval;
#endif
				break;
			case 3:
				suspend_default_console_level = nextval;
#ifndef CONFIG_SOFTWARE_SUSPEND_DEBUG
				if (suspend_default_console_level > 1)
					suspend_default_console_level = 1;
#endif
				break;
			case 4:
				image_size_limit = nextval;
				break;
			case 5:
				max_async_ios = nextval;
				if (max_async_ios > 256)
					max_async_ios = 256;
				break;

		}
		buf++;
		while (*buf == ' ')
			buf++;
	}
	buf[count] = 0;
	free_pages((unsigned long) buf, 0);
	return count;
}

/*
 * suspend_read_compat_proc.
 *
 * Like it's _write_ sibling, this entry allows all of the settings
 * to be read at once.
 * It too was originally for compatibility with pre- /proc/suspend
 * versions, but has been retained because it makes saving and
 * restoring the configuration simpler.
 */
static int suspend_read_compat_proc(char * page, char ** start, off_t off, int count,
		int *eof, void *data)
{
	int len = 0;

	len = sprintf(page, "%ld %ld %ld %d %d %d\n",
			suspend_result,
			suspend_action,
			suspend_debug_state,
			suspend_default_console_level,
			image_size_limit,
			max_async_ios);
	*eof = 1;
	return len;
}

/*
 * proc_software_suspend_pending
 *
 * This routine initiates a suspend cycle when /proc/suspend/activity is
 * written to. The value written is ignored.
 */

static int proc_software_suspend_pending(struct file *file, const char *buffer,
        unsigned long count, void *data)
{
	software_suspend_pending();
	return count;
}

extern int attempt_to_parse_resume_device(int boot_time);

extern void start_ksuspendd(void * data);
extern int initialise_suspend_plugins(void);
extern void cleanup_suspend_plugins(void);

static int resume2_write_proc(void)
{
	mm_segment_t	oldfs;

	oldfs = get_fs(); set_fs(KERNEL_DS);
	initialise_suspend_plugins();
	attempt_to_parse_resume_device(0);
	cleanup_suspend_plugins();
	set_fs(oldfs);
	return 0;
}

extern int debuginfo_read_proc(char * page, char ** start, off_t off, int count,
		int *eof, void *data);
/*
 * generic_read_proc
 *
 * Generic handling for reading the contents of bits, integers,
 * unsigned longs and strings.
 */
static int generic_read_proc(char * page, char ** start, off_t off, int count,
		int *eof, void *data)
{
	int len = 0;
	struct suspend_proc_data * proc_data = (struct suspend_proc_data *) data;

	switch (proc_data->type) {
		case SWSUSP_PROC_DATA_CUSTOM:
			printk("Error! /proc/suspend/%s marked as having custom"
				" routines, but the generic read routine has"
				" been invoked.\n",
				proc_data->filename);
			break;
		case SWSUSP_PROC_DATA_BIT:
			len = sprintf(page, "%d\n", 
				-test_bit(proc_data->data.bit.bit,
					proc_data->data.bit.bit_vector));
			break;
		case SWSUSP_PROC_DATA_INTEGER:
			{
				int * variable = proc_data->data.integer.variable;
				len = sprintf(page, "%d\n", *variable);
				break;
			}
		case SWSUSP_PROC_DATA_UL:
			{
				long * variable = proc_data->data.ul.variable;
				len = sprintf(page, "%lu\n", *variable);
				break;
			}
		case SWSUSP_PROC_DATA_STRING:
			{
				char * variable = proc_data->data.string.variable;
				len = sprintf(page, "%s\n", variable);
				break;
			}
	}
	*eof = 1;
	return len;
}

/*
 * generic_write_proc
 *
 * Generic routine for handling writing to files representing
 * bits, integers and unsigned longs.
 */

static int generic_write_proc(struct file *file, const char * buffer,
		unsigned long count, void * data)
{
	struct suspend_proc_data * proc_data = (struct suspend_proc_data *) data;
	char * my_buf = (char *) get_zeroed_page(GFP_ATOMIC);
	int result = count;

	if (!my_buf)
		return -ENOMEM;

	if (count > PAGE_SIZE)
		count = PAGE_SIZE;

	if (copy_from_user(my_buf, buffer, count))
		return -EFAULT;
	
	my_buf[count] = 0;

	switch (proc_data->type) {
		case SWSUSP_PROC_DATA_CUSTOM:
			printk("Error! /proc/suspend/%s marked as having custom"
				" routines, but the generic write routine has"
				" been invoked.\n",
				proc_data->filename);
			break;
		case SWSUSP_PROC_DATA_BIT:
			{
			int value = simple_strtoul(my_buf, NULL, 0);
			if (value)
				set_bit(proc_data->data.bit.bit, 
					(proc_data->data.bit.bit_vector));
			else
				clear_bit(proc_data->data.bit.bit,
					(proc_data->data.bit.bit_vector));
			}
			break;
		case SWSUSP_PROC_DATA_INTEGER:
			{
				int * variable = proc_data->data.integer.variable;
				int minimum = proc_data->data.integer.minimum;
				int maximum = proc_data->data.integer.maximum;
				*variable = simple_strtol(my_buf, NULL, 0);
				if (((*variable) < minimum))
					*variable = minimum;

				if (((*variable) > maximum))
					*variable = maximum;
				break;
			}
		case SWSUSP_PROC_DATA_UL:
			{
				unsigned long * variable = proc_data->data.ul.variable;
				unsigned long minimum = proc_data->data.ul.minimum;
				unsigned long maximum = proc_data->data.ul.maximum;
				*variable = simple_strtoul(my_buf, NULL, 0);
				
				if ((*variable) < minimum)
					*variable = minimum;

				if ((*variable) > maximum)
					*variable = maximum;
				break;
			}
			break;
		case SWSUSP_PROC_DATA_STRING:
			{
				int copy_len = 
					(count > 
					 proc_data->data.string.max_length) ?
					proc_data->data.string.max_length : 
					count;
				char * variable =
					proc_data->data.string.variable;
				strncpy(variable, my_buf, copy_len);
				if ((copy_len) &&
					 (my_buf[copy_len - 1] == '\n'))
					variable[count - 1] = 0;
				variable[count] = 0;
				if (proc_data->data.string.write_proc) {
					int routine_result = 
					  proc_data->data.string.write_proc();
					if (routine_result < 0)
						result = routine_result;
				}
			}
			break;
	}
	free_pages((unsigned long) my_buf, 0);
	return result;
}

/*
 * Non-plugin proc entries.
 *
 * This array contains entries that are automatically registered at
 * boot. Plugins and the console code register their own entries separately.
 */

static struct suspend_proc_data proc_params[] = {
	{ .filename			= "activate",
	  .permissions			= PROC_WRITEONLY,
	  .type				= SWSUSP_PROC_DATA_CUSTOM,
	  .data = {
		  .special = {
			  .write_proc	= proc_software_suspend_pending
		  }
	  }
	},

	{ .filename			= "all_settings",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_CUSTOM,
	  .data = {
		  .special = {
			  .read_proc	= suspend_read_compat_proc,
			  .write_proc	= suspend_write_compat_proc,
		  }
	  }
	},

	{ .filename			= "async_io_limit", 
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_INTEGER,
	  .data = {
		  .integer = {
			  .variable	= &max_async_ios,
			  .minimum	= 1,
			  .maximum	= 256,
		  }
	  }
	},
		
	{ .filename			= "debug_info",
	  .permissions			= PROC_READONLY,
	  .type				= SWSUSP_PROC_DATA_CUSTOM,
	  .data = {
		  .special = {
			  .read_proc	= debuginfo_read_proc,
		  }
	  }
	},
	
	{ .filename			= "image_size_limit",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_INTEGER,
	  .data = {
		  .integer = {
			  .variable	= &image_size_limit,
			  .minimum	= -2,
			  .maximum	= 32767,
		  }
	  }
	},

	{ .filename			= "interface_version",
	  .permissions			= PROC_READONLY,
	  .type				= SWSUSP_PROC_DATA_INTEGER,
	  .data = {
		  .integer = {
			  .variable	= &suspend_proc_version,
		  }
	  }
	},
	
	{ .filename			= "last_result",
	  .permissions			= PROC_READONLY,
	  .type				= SWSUSP_PROC_DATA_UL,
	  .data = {
		  .ul = {
			  .variable	=  &suspend_result,
		  }
	  }
	},
	
	{ .filename			= "reboot",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &suspend_action,
			  .bit		= SUSPEND_REBOOT,
		  }
	  }
	},
	  
	{ .filename			= "resume2",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_STRING,
	  .data = {
		  .string = {
			  .variable	= resume_file,
			  .max_length	= 255,
			  .write_proc	= resume2_write_proc,
		  }
	  }
	},


	{ .filename			= "version",
	  .permissions			= PROC_READONLY,
	  .type				= SWSUSP_PROC_DATA_STRING,
	  .data = {
		  .string = {
			  .variable	= suspend_core_version,
		  }
	  }
	},

#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	{ .filename			= "freezer_test",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &suspend_action,
			  .bit		= SUSPEND_FREEZER_TEST,
		  }
	  }
	},

	{ .filename			= "slow",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &suspend_action,
			  .bit		= SUSPEND_SLOW,
		  }
	  }
	},
#endif
	  
#ifdef CONFIG_SOFTWARE_SUSPEND_KEEP_IMAGE
	{ .filename			= "keep_image",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &suspend_action,
			  .bit		= SUSPEND_KEEP_IMAGE,
		  }
	  }
	},
#endif
};

/*
 * suspend_initialise_proc
 *
 * Initialise the /proc/suspend tree.
 *
 */

int suspend_initialise_proc(void)
{
	suspend_dir = proc_mkdir("swsusp", NULL);
	
	if (suspend_dir == NULL) {
		printk(KERN_ERR "Failed to create /proc/swsusp.\n");
		return 1;
	}

	INIT_LIST_HEAD(&suspend_proc_entries);

	proc_initialised = 1;

	return 0;
}

/*
 * suspend_register_procfile
 *
 * Helper for registering a new /proc/suspend entry.
 */

struct proc_dir_entry * suspend_register_procfile(
		struct suspend_proc_data * suspend_proc_data)
{
	struct proc_dir_entry * new_entry;
	
	if ((!proc_initialised) && (suspend_initialise_proc()))
		return NULL;

	new_entry = create_proc_entry(
			suspend_proc_data->filename,
			suspend_proc_data->permissions, 
			suspend_dir);
	if (new_entry) {
		list_add_tail(&suspend_proc_data->proc_data_list, &suspend_proc_entries);
		if (suspend_proc_data->type) {
			new_entry->read_proc = generic_read_proc;
			new_entry->write_proc = generic_write_proc;
		} else {
			new_entry->read_proc = suspend_proc_data->data.special.read_proc;
			new_entry->write_proc = suspend_proc_data->data.special.write_proc;
		}
		new_entry->data = suspend_proc_data;
	}
	return new_entry;
}

/*
 * suspend_unregister_procfile
 *
 * Helper for removing unwanted /proc/suspend entries.
 *
 */
void suspend_unregister_procfile(struct suspend_proc_data * suspend_proc_data)
{
	remove_proc_entry(
		suspend_proc_data->filename,
		suspend_dir);
	list_del(&suspend_proc_data->proc_data_list);
}

/*
 * suspend_init_proc
 * 
 * Initialise proc file support at boot time. This is called
 * by suspend2.c::software_resume2.
 */
int suspend_init_proc(void)
{
	int i;
	int numfiles = sizeof(proc_params) / sizeof(struct suspend_proc_data);
	
	if ((!proc_initialised) && suspend_initialise_proc())
		return 1;

	for (i=0; i< numfiles; i++)
		suspend_register_procfile(&proc_params[i]);

	suspend_console_proc_init();

	compat_parent = find_proc_dir_entry("sys", &proc_root);
	if (compat_parent) {
		compat_parent = find_proc_dir_entry("kernel", compat_parent);
		if (compat_parent) {
			compat_entry = create_proc_entry("swsusp",
				0600, 
				compat_parent);
			compat_entry->read_proc = suspend_read_compat_proc;
			compat_entry->write_proc = suspend_write_compat_proc;
		}
	}

	return 0;
}
