/*
 * kernel/power/suspend2.c
 *
 * Copyright (C) 1998-2001 Gabor Kuti <seasons@fornax.hu>
 * Copyright (C) 1998,2001,2002 Pavel Machek <pavel@suse.cz>
 * Copyright (C) 2002-2003 Florent Chabaud <fchabaud@free.fr>
 * Copyright (C) 2002-2004 Nigel Cunningham <ncunningham@users.sourceforge.net>
 *
 * This file is released under the GPLv2.
 *
 * This file is to realize architecture-independent
 * machine suspend feature using pretty near only high-level routines
 *
 * We'd like to thank the following people for their work:
 * 
 * Pavel Machek <pavel@ucw.cz>:
 * Modifications, defectiveness pointing, being with me at the very beginning,
 * suspend to swap space, stop all tasks. Port to 2.4.18-ac and 2.5.17.
 *
 * Steve Doddi <dirk@loth.demon.co.uk>: 
 * Support the possibility of hardware state restoring.
 *
 * Raph <grey.havens@earthling.net>:
 * Support for preserving states of network devices and virtual console
 * (including X and svgatextmode)
 *
 * Kurt Garloff <garloff@suse.de>:
 * Straightened the critical function in order to prevent compilers from
 * playing tricks with local variables.
 *
 * Andreas Mohr <a.mohr@mailto.de>
 *
 * Alex Badea <vampire@go.ro>:
 * Fixed runaway init
 *
 * Jeff Snyder <je4d@pobox.com>
 * ACPI patch
 *
 * Nathan Friess <natmanz@shaw.ca>
 * Some patches.
 *
 * Michael Frank <mhf@linuxmail.org>
 * Extensive testing and help with improving stability.
 *
 * More state savers are welcome. Especially for the scsi layer...
 *
 * For TODOs,FIXMEs also look in Documentation/suspend.txt.
 * 
 * Variable definitions which are needed if PM is enabled but 
 * SOFTWARE_SUSPEND is disabled are found near the top of process.c.
 */

#define SWSUSP_MAIN_C

#include <linux/suspend-common.h>
#include <linux/suspend-version-specific.h>
#ifdef  CONFIG_X86
#include <asm/i387.h> /* for kernel_fpu_end */
#endif
#ifdef CONFIG_KDB
#include <linux/kdb.h>
#include <linux/kallsyms.h>
#endif

#if (SWSUSP_VERSION_SPECIFIC_REVISION != SWSUSP_CORE_REVISION)
#error The core and version specific patch revisions of Software Suspend are
#error incompatible. You can find your core version from
#error include/linux/suspend-debug.h and your version specific revision from
#error include/linux/suspend-version-specific.h.
#endif

#include <linux/console.h>

unsigned int nr_suspends = 0;

/* Variables to be preserved over suspend */
int pageset1_sizelow = 0, pageset2_sizelow = 0;
struct pagedir __nosavedata pagedir_resume;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
static int pm_suspend_state = 0;
#endif

struct list_head suspend_filters, suspend_writers, suspend_plugins;
int num_filters = 0, num_writers = 0;
struct suspend_plugin_ops * active_writer = NULL;

int suspend_act_used = 0;
int suspend_lvl_used = 0;
int suspend_dbg_used = 0;

/* For resume2= kernel option */
char resume_file[256] = CONFIG_SOFTWARE_SUSPEND_DEFAULT_RESUME2;

unsigned long orig_mem_free = 0;

static void drivers_resume(int);
extern void do_suspend2_lowlevel(int resume);
extern unsigned long header_storage_for_plugins(void);
extern int pm_prepare_console(void);
extern void pm_restore_console(void);
extern void suspend_relinquish_console(void);
extern int suspend_io_time[2][2];
void empty_suspend_memory_pool(void);
int read_primary_suspend_image(void);
extern void display_nosave_pages(void);
int initialise_suspend_plugins(void);
void cleanup_suspend_plugins(void);
int attempt_to_parse_resume_device(int boot_time);

unsigned long * in_use_map = NULL;
unsigned long * pageset2_map = NULL;
unsigned long * checksum_map = NULL;

/* Suspend pagedir is allocated before final copy, and saved with data
 * as well as separately, so it can be used to load pageset2 at resume.
 *
 * At resume, the original pagedir is loaded as pagedir_resume and then
 * moved to a place where it doesn't collide with the data to be copied
 * back. Then the data is read (again into spots that don't collide) and
 * then copied back, giving us access to the saved pagedir again and
 * forgetting the loaded pagedir at the same time. We then used the saved
 * pagedir to load pageset2 (if necessary) before freeing that pagedir.
 */

struct pagedir pagedir1 = { 1, 0, 0}, pagedir2 = {2, 0, 0};

#ifdef CONFIG_SOFTWARE_SUSPEND_VARIATION_ANALYSIS

#define CHECKSUMS_PER_PAGE ((PAGE_SIZE - sizeof(void *)) / sizeof(unsigned long))
#define NEXT_CHECKSUM_PAGE(page) *((unsigned long *) (((char *) (page)) + PAGE_SIZE - sizeof(void *)))
static int checksum_pages;
static unsigned long * first_checksum_page, *last_checksum_page;
static struct reload_data * first_reload_data, * last_reload_data;

static void suspend_print_differences(void);
void suspend_calculate_checksums(void);
void suspend_check_checksums(void);
#endif

unsigned long suspend_debug_state = 0;
char * debug_info_buffer;

int image_size_limit = 0;
int max_async_ios = 32;

/* Pagedir.c */
extern void copy_pageset1(void);
extern int allocatemap(unsigned long ** pagemap, int setnosave);
extern void free_pagedir(struct pagedir * p);
extern int freemap(unsigned long ** pagemap);

/* Prepare_image.c */

extern int prepare_image(void);

/* proc.c */

extern int suspend_init_proc(void);
extern int suspend_cleanup_proc(void);

/* process.c */
extern atomic_t suspend_cpu_counter; 

/* include/asm-i386/suspend.h */
extern void smp_suspend2_lowlevel(void * info);

static void ensure_on_processor_zero(void)
{
#ifdef CONFIG_SMP
	set_cpus_allowed(current, CPU0_MASK);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	if (unlikely(cpu_number_map(smp_processor_id()) != 0))
		BUG();
#else
	BUG_ON(smp_processor_id() != 0);
#endif
#endif
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
/* Make disk drivers accept operations, again */
void storage_unsuspend(void)
{
#ifdef CONFIG_SCSI
	struct pm_dev *dev = NULL;

	while ((dev = pm_find(PM_SCSI_DEV, dev)))
		pm_send(dev, PM_RESUME, (void *) 0);

#endif
#ifdef CONFIG_BLK_DEV_IDE
	ide_disk_unsuspend(0);
#endif
}

void storage_suspend(void)
{
#ifdef CONFIG_SCSI
	struct pm_dev *dev = NULL;

	while ((dev = pm_find(PM_SCSI_DEV, dev)))
		pm_send(dev, PM_SUSPEND, (void *) 3);

#endif
#ifdef CONFIG_BLK_DEV_IDE
	do_suspend_sync();

	ide_disk_suspend();
#endif
}
#else
void storage_unsuspend(void)
{
#if 0 /* Work in progress */
	int result;

	result = device_resume_type(SCSI);
	return result;
#endif
}

void storage_suspend(void)
{
	int result;

	result = device_suspend(3);
}
#endif

#define RESUME_PHASE1 1 /* Called from interrupts disabled */
#define RESUME_PHASE2 2 /* Called with interrupts enabled */
#define RESUME_ALL_PHASES (RESUME_PHASE1 | RESUME_PHASE2)

void drivers_resume(int flags)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)

#ifdef CONFIG_BLK_DEV_HD
  	if(flags & RESUME_PHASE2)
		do_reset_hd();	/* Kill all controller state */
#endif

  	if (flags & RESUME_PHASE1) {
#ifdef CONFIG_BLK_DEV_IDE
		ide_disk_unsuspend(1);		
#endif
#ifdef CONFIG_SCSI
		{
		struct pm_dev *dev = NULL;

		while ((dev = pm_find(PM_SCSI_DEV, dev)))
			pm_send(dev, PM_RESUME, (void *) 0);
		}
#endif
#ifdef CONFIG_BLK_DEV_MD
		md_autostart_arrays();
#endif
	}

  	if (flags & RESUME_PHASE2) {
		if (pm_suspend_state) {
			if (pm_send_all(PM_RESUME,(void *)0))
				printk(name_suspend
				  "Problem while sending resume event\n");
			pm_suspend_state=0;
		} else
			printk(name_suspend "PM suspend state wasn't raised\n");

#ifdef DEFAULT_SUSPEND_CONSOLE
		update_screen(fg_console);
#endif
	}

#else /* Kernel version >= 2.5.0 */
	if (flags & RESUME_PHASE1)
		device_resume();
#endif
}

/* Called from process context */
static int drivers_suspend(void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#ifdef CONFIG_BLK_DEV_MD
	md_notify_reboot(NULL, SYS_HALT, NULL);
#endif
	STORAGE_SUSPEND
	if (!pm_suspend_state) {
		if (pm_send_all(PM_SUSPEND,(void *)3)) {
			printk(name_suspend
				"Problem while sending suspend event\n");
			drivers_resume(RESUME_ALL_PHASES);
			return(1);
		}
		pm_suspend_state=1;
	}
	return(0);
#else
	int result;

	result = device_suspend(3);
	return result;
#endif
}

#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG

int suspend_free_mem_values[MAX_FREEMEM_SLOTS][2];
/* These should match the enumerated type in suspend-common.h */
static char * suspend_free_mem_descns[MAX_FREEMEM_SLOTS] = {
	"Start/End      ", /* 0 */
	"Console Allocn ",
	"Drain pcp      ",
	"InUse map      ",
	"PS2 map        ",
	"Checksum map   ", /* 5 */
	"Reload pages   ",
	"Init plugins   ",
	"Memory pool    ",
	"Freezer        ",
	"Eat Memory     ", /* 10 */
	"Syncing        ",
	"Grabbed Memory ",
	"Range Pages    ",
	"Extra PD1 pages",
	"Writer storage ", /* 15 */
	"Header storage ",
	"Checksum pages ",
	"KStat data     ",
	"Debug Info     ",
	"Remove Image   ", /* 20 */
	"I/O            ",
	"I/O info       ",
	"Start one      ",
};

/* store_free_mem
 */


void suspend_store_free_mem(int slot, int side)
{
	static int last_free_mem;
	int this_free_mem = nr_free_pages() + suspend_amount_grabbed +
			suspend_memory_pool_level();
	int i;

	BUG_ON(slot >= MAX_FREEMEM_SLOTS);

	printnolog(SUSPEND_MEMORY, SUSPEND_VERBOSE, 0,
			"Last free mem was %d. Is now %d. ",
			last_free_mem, this_free_mem);

	if (slot == 0) {
		if (!side)
			for (i = 1; i < MAX_FREEMEM_SLOTS; i++) {
				suspend_free_mem_values[i][0] = 0;
				suspend_free_mem_values[i][1] = 0;
			}
		suspend_free_mem_values[slot][side] = this_free_mem;
	} else
		suspend_free_mem_values[slot][side] += this_free_mem - last_free_mem;
	last_free_mem = this_free_mem;
	printnolog(SUSPEND_MEMORY, SUSPEND_VERBOSE, 0,
		"%s value %d now %d.\n",
		suspend_free_mem_descns[slot],
		side,
		suspend_free_mem_values[slot][side]);
}

/*
 * display_free_mem
 */
static void display_free_mem(void)
{
	int i;


	if (!TEST_DEBUG_STATE(SUSPEND_MEMORY))
		return;
	
	printnolog(SUSPEND_MEMORY, SUSPEND_HIGH, 0,
		"Start:       %7d    End: %7d.\n",
		suspend_free_mem_values[0][0],
		suspend_free_mem_values[0][1]);
	
	for (i = 1; i < MAX_FREEMEM_SLOTS; i++)
		if (suspend_free_mem_values[i][0] + suspend_free_mem_values[i][1])
			printnolog(SUSPEND_MEMORY, SUSPEND_HIGH, 0,
				"%s %7d         %7d.\n",
				suspend_free_mem_descns[i],
				suspend_free_mem_values[i][0],
				suspend_free_mem_values[i][1]);
}
#endif

/*
 * save_image
 * Result code (int): Zero on success, non zero on failure.
 * Functionality    : High level routine which performs the steps necessary
 *                    to prepare and save the image after preparatory steps
 *                    have been taken.
 * Key Assumptions  : Processes frozen, sufficient memory available, drivers
 *                    suspended.
 * Called from      : do_suspend2_suspend_2
 */
extern struct pageset_sizes_result recalculate_stats(void);
extern void display_stats(void);
extern int write_pageset(struct pagedir * pagedir, int whichtowrite);
extern int write_image_header(void);
extern int read_secondary_pagedir(int overwrittenpagesonly);

static int save_image(void)
{
	int temp_result;

	if (RAM_TO_SUSPEND > max_mapnr) {
		prepare_status(1, 1,
			"Couldn't get enough free pages, on %ld pages short",
			 RAM_TO_SUSPEND - max_mapnr);
		goto abort_saving;
	}
	
	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,
		" - Final values: %d and %d.\n",
		pageset1_size, 
		pageset2_size);

	pagedir1.lastpageset_size = pageset1_size;
	pagedir2.lastpageset_size = pageset2_size;

	check_shift_keys(1, "About to write pagedir2.");

	temp_result = write_pageset(&pagedir2, 2);
	
	if (temp_result == -1 || TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto abort_saving;

	check_shift_keys(1, "About to copy pageset 1.");

	drivers_suspend();

	PRINTPREEMPTCOUNT("Prior to lowlevel suspend.");
	
	do_suspend2_lowlevel(0);

	PRINTPREEMPTCOUNT("At exit from save_image.");
	return 0;
abort_saving:
	PRINTPREEMPTCOUNT("At exit from save_image.");
	return -1;		
}

int save_image_part1(void)
{
	int temp_result;

	PRINTPREEMPTCOUNT("Prior to copying pageset 1.");

	/* Set the progress bar to what we want at resume time, so that
	 * it looks smoother then */
	update_status(pageset1_size, pageset1_size + pageset2_size, NULL);
	
	copy_pageset1();
	
	update_status(pageset2_size, pageset1_size + pageset2_size, NULL);
	
	/*
	 *  ----   FROM HERE ON, NEED TO REREAD PAGESET2 IF ABORTING!!! -----
	 *  
	 */
	
	spin_unlock_irqrestore(&suspend_irq_lock, suspendirqflags);
	
#ifdef CONFIG_PREEMPT
	preempt_enable();
#endif

#ifdef CONFIG_X86
	kernel_fpu_end();
#endif

	PRINTPREEMPTCOUNT("Prior to resuming drivers.");

	drivers_resume(RESUME_PHASE1);
	
	STORAGE_UNSUSPEND

	/* 
	 * Other processors have waited for me to make the atomic copy of the 
	 * kernel
	 */

	software_suspend_state &= ~SOFTWARE_SUSPEND_FREEZE_SMP;

	while (atomic_read(&suspend_cpu_counter) &&
		(!TEST_RESULT_STATE(SUSPEND_ABORTED))) {
		cpu_relax();
		barrier();
		check_shift_keys(0, "");
	}

	if (TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto abort_reloading_pagedir_two;

	check_shift_keys(1, "About to write pageset1.");

	/*
	 * End of critical section.
	 */
	
	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,"-- Writing pageset1\n");

	temp_result = write_pageset(&pagedir1, 1);

	if (temp_result == -1 || TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto abort_reloading_pagedir_two;

	check_shift_keys(1, "About to write header.");

	if (TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto abort_reloading_pagedir_two;

	temp_result = write_image_header();

	if (temp_result || (TEST_RESULT_STATE(SUSPEND_ABORTED)))
		goto abort_reloading_pagedir_two;

	check_shift_keys(1, "About to power down or reboot.");

	PRINTPREEMPTCOUNT("At exit from save_image.");
	return 0;

abort_reloading_pagedir_two:
	temp_result = read_secondary_pagedir(1);

	/* If that failed, we're sunk. Panic! */
	if (temp_result)
		panic("Attempt to reload pagedir 2 while aborting "
				"a suspend failed.");

	PRINTPREEMPTCOUNT("At exit from save_image.");
	return -1;		

}

/*
 * suspend_power_down
 * Functionality   : Powers down or reboots the computer once the image
 *                   has been written to disk.
 * Key Assumptions : Able to reboot/power down via code called or that
 *                   the warning emitted if the calls fail will be visible
 *                   to the user (ie printk resumes devices).
 * Called From     : do_suspend2_suspend_2
 */

extern asmlinkage long sys_reboot(int magic1, int magic2, unsigned int cmd,
	       void * arg);
extern void apm_power_off(void);

void suspend_power_down(void)
{
	C_A_D = 0;
	/* No delay should be needed - ide_disk_suspend purges cache now */
	prepare_status(1, 0, "Ready to power down.");
#if defined(CONFIG_VT)
	if (TEST_ACTION_STATE(SUSPEND_REBOOT))
		sys_reboot(LINUX_REBOOT_MAGIC1, LINUX_REBOOT_MAGIC2,
				LINUX_REBOOT_CMD_RESTART, NULL);
#endif
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
	drivers_suspend();
#endif
	sys_reboot(LINUX_REBOOT_MAGIC1, LINUX_REBOOT_MAGIC2,
			LINUX_REBOOT_CMD_POWER_OFF, NULL);

	prepare_status(1, 0, "Probably not capable for powerdown.");
	machine_halt();
	prepare_status(1, 0, "System is now halted.");
	while (1)
		cpu_relax();
	/* NOTREACHED */
}

/*
 * do_suspend2_resume_1
 * Functionality   : Preparatory steps for copying the original kernel back.
 * Called From     : include/asm/suspend.h:do_suspend2_lowlevel
 */

void do_suspend2_resume_1(void)
{
	drivers_suspend();

	/* Get other cpus ready to restore their original contexts */

	software_suspend_state |= SOFTWARE_SUSPEND_FREEZE_SMP;

	smp_call_function(smp_suspend2_lowlevel, NULL, 0, 0);

	while (atomic_read(&suspend_cpu_counter) < (NUM_CPUS - 1)) {
		cpu_relax();
		barrier();
	}

	barrier();
	mb();
	/* Done to disable interrupts */ 
	spin_lock_irqsave(&suspend_irq_lock, suspendirqflags);

	printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,
			name_suspend "About to copy pageset1 back...\n");

	MDELAY(2000);
}

/*
 * do_suspend2_resume_2
 * Functionality   : Steps taken after copying back the original kernel at
 *                   resume.
 * Key Assumptions : Will be able to read back secondary pagedir (if 
 *                   applicable).
 * Called From     : include/asm/suspend.h:do_suspend2_lowlevel
 */

extern void post_resume_console_redraw(void);
extern int suspend_reread_pages(struct reload_data * page_list);

void do_suspend2_resume_2(void)
{
	tainted |= 4; /* Taint the kernel because we have suspended */
	now_resuming = 1;
	software_suspend_state |= SOFTWARE_SUSPEND_PAGESET2_NOT_LOADED;

	PRINTPREEMPTCOUNT("In resume_2 after copy back.");

#ifdef CONFIG_PREEMPT
	preempt_enable();
#endif

	spin_unlock_irqrestore(&suspend_irq_lock, suspendirqflags);

#ifdef DEFAULT_SUSPEND_CONSOLE
	post_resume_console_redraw();
#endif

	check_shift_keys(1, "Post resume");

	prepare_status(0, 0, "Post kernel restoration.");

	drivers_resume(RESUME_PHASE1);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	STORAGE_UNSUSPEND
#endif
	
	check_shift_keys(1, "About to reload secondary pagedir.");

	read_secondary_pagedir(0);
	software_suspend_state &= ~SOFTWARE_SUSPEND_PAGESET2_NOT_LOADED;
	
#ifdef CONFIG_SOFTWARE_SUSPEND_VARIATION_ANALYSIS
	if (CHECKMASK(SUSPEND_INTEGRITY)) {
		suspend_reread_pages(first_reload_data);
		suspend_print_differences();
	}
#endif

	prepare_status(0, 0, "Cleaning up...");

#ifdef CONFIG_SOFTWARE_SUSPEND_KEEP_IMAGE
	if (TEST_ACTION_STATE(SUSPEND_KEEP_IMAGE)) {
		printlog(SUSPEND_ANY_SECTION, SUSPEND_LOW,
			name_suspend "Not invalidating the image due "
			"to Keep Image being enabled.\n");
		SET_RESULT_STATE(SUSPEND_KEPT_IMAGE);
	} else
#endif
		active_writer->ops.writer.invalidate_image();

	software_suspend_state &= ~SOFTWARE_SUSPEND_USE_MEMORY_POOL;
	
	drivers_resume(RESUME_ALL_PHASES);
}

/*
 * do_suspend2_suspend_1
 * Functionality   : Steps taken prior to saving CPU state and the image
 *                   itself.
 * Called From     : include/asm/suspend.h:do_suspend2_lowlevel
 */

void do_suspend2_suspend_1(void)
{
	/* Save other cpu contexts */

	software_suspend_state |= SOFTWARE_SUSPEND_FREEZE_SMP;

	smp_call_function(smp_suspend2_lowlevel, NULL, 0, 0);

	while ((atomic_read(&suspend_cpu_counter) < (NUM_CPUS - 1)) &&
		(!TEST_RESULT_STATE(SUSPEND_ABORTED))) {
		cpu_relax();
		barrier();
		check_shift_keys(0, "");
	}

	mb();
	barrier();

#ifdef CONFIG_PREEMPT
	preempt_disable();
#endif
	spin_lock_irqsave(&suspend_irq_lock, suspendirqflags);
}

/*
 * do_suspend2_suspend_2
 * Functionality   : Steps taken after saving CPU state to save the
 *                   image and powerdown/reboot or recover on failure.
 * Key Assumptions : save_image returns zero on success; otherwise we need to
 *                   clean up and exit. The state on exiting this routine 
 *                   should be essentially the same as if we have suspended,
 *                   resumed and reached the end of do_suspend2_resume_2.
 * Called From     : include/asm/suspend.h:do_suspend2_lowlevel
 */
void do_suspend2_suspend_2(void)
{
	if (!save_image_part1())
		suspend_power_down();

	if (!TEST_RESULT_STATE(SUSPEND_ABORT_REQUESTED))
		printk(KERN_EMERG name_suspend
			"Suspend failed, trying to recover...\n");
	MDELAY(1000);

	barrier();
	mb();

	active_writer->ops.writer.invalidate_image();
	drivers_resume(RESUME_ALL_PHASES);
}

static inline void lru_check_page(struct page * page)
{
	if (!PageLRU(page))
		printk("Page %p/%p in inactivelist but not marked LRU.\n",
			page, page_address(page));
}

/*
 * print_plugin_debug_info
 * Functionality   : Get debugging info from plugins into a buffer.
 */
static int print_plugin_debug_info(char * buffer, int buffer_size)
{
	struct list_head *plugin;
	struct suspend_plugin_ops *this_plugin = NULL;
	int len = 0;

	list_for_each(plugin, &suspend_plugins) {
		this_plugin = list_entry(plugin, struct suspend_plugin_ops,
				plugin_list);
		if (this_plugin->disabled)
			continue;
		if (this_plugin->print_debug_info) {
			int result;
			result = this_plugin->print_debug_info(buffer + len, 
					buffer_size - len);
			len += result;
		}
	}

	return len;
}

/* get_debug_info
 * Functionality:	Store debug info in a buffer.
 * Called from:		Worker thread or via software_suspend_pending.
 */

#define SNPRINTF(a...) 	len += suspend_snprintf(debug_info_buffer + len, \
		PAGE_SIZE - len - 1, ## a);

static int get_suspend_debug_info(void)
{
	int len = 0;
	if (!debug_info_buffer) {
		debug_info_buffer = (char *) get_zeroed_page(GFP_ATOMIC);
		if (!debug_info_buffer) {
			printk("Error! Unable to allocate buffer for"
					"software suspend debug info.\n");
			return 0;
		}
	}

	SNPRINTF("Please include the following information in bug reports:\n");
	SNPRINTF("- SWSUSP core    : %s\n", SWSUSP_CORE_VERSION);
	SNPRINTF("- Kernel Version : %s\n", UTS_RELEASE);
	SNPRINTF("- Version spec.  : %s\n",
			SWSUSP_VERSION_SPECIFIC_REVISION_STRING);
	SNPRINTF("- Compiler vers. : %d.%d\n", __GNUC__, __GNUC_MINOR__);
#ifdef CONFIG_MODULES
	SNPRINTF("- Modules loaded : ");
	{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
		struct module *this_mod;
		this_mod = module_list;
		while (this_mod) {
			if (this_mod->name)
				SNPRINTF("%s ", this_mod->name);
			this_mod = this_mod->next;
		}
#else
		extern int print_module_list_to_buffer(char * buffer, int size);
		len+= print_module_list_to_buffer(debug_info_buffer + len,
				PAGE_SIZE - len - 1);
#endif
	}
	SNPRINTF("\n");
#else
	SNPRINTF("- No module support.\n");
#endif
	SNPRINTF("- Attempt number : %d\n", nr_suspends);
	if (num_range_pages)
		SNPRINTF("- Pageset sizes  : %d and %d (%d low).\n",
			pagedir1.lastpageset_size, 
			pagedir2.lastpageset_size, 
			pageset2_sizelow);
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	SNPRINTF("- Parameters     : %ld %ld %ld %d %d %d\n",
			suspend_result,
			suspend_action,
			suspend_debug_state,
			suspend_default_console_level,
			image_size_limit,
			max_async_ios);
#else
	SNPRINTF("- Parameters     : %ld %ld %d %d\n",
			suspend_result,
			suspend_action,
			image_size_limit,
			max_async_ios);
#endif
	if (num_range_pages)
		SNPRINTF("- Calculations   : Image size: %lu. "
			"Ram to suspend: %ld.\n",
			STORAGE_NEEDED(1), RAM_TO_SUSPEND);
	SNPRINTF("- Limits         : %lu pages RAM. Initial boot: %lu.\n",
		max_mapnr, orig_mem_free);
	SNPRINTF("- Overall expected compression percentage: %d.\n",
			100 - expected_compression_ratio());
	len+= print_plugin_debug_info(debug_info_buffer + len, 
			PAGE_SIZE - len - 1);
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	SNPRINTF("- Debugging compiled in.\n");
#endif
#ifdef CONFIG_PREEMPT
	SNPRINTF("- Preemptive kernel.\n");
#endif
#ifdef CONFIG_SMP
	SNPRINTF("- SMP kernel.\n");
#endif
#ifdef CONFIG_HIGHMEM
	SNPRINTF("- Highmem Support.\n");
#endif
	if (num_range_pages)
		SNPRINTF("- Max ranges used: %d ranges in %d pages.\n",
			max_ranges_used, num_range_pages);
	if (suspend_io_time[0][1] && suspend_io_time[1][1])
		SNPRINTF("- I/O speed: Write %d MB/s, Read %d MB/s.\n",
			(MB((unsigned long) suspend_io_time[0][0]) * HZ /
			 suspend_io_time[0][1]),
			(MB((unsigned long) suspend_io_time[1][0]) * HZ /
			 suspend_io_time[1][1]))
	else if (num_range_pages)
		SNPRINTF("- Suspend cancelled. No I/O speed stats.\n");

	return len;
}

/*
 * debuginfo_read_proc
 * Functionality   : Displays information that may be helpful in debugging
 *                   software suspend.
 */
int debuginfo_read_proc(char * page, char ** start, off_t off, int count,
		int *eof, void *data)
{
	int info_len, result;

	initialise_suspend_plugins();
	info_len = get_suspend_debug_info();
	cleanup_suspend_plugins();

	memcpy(page, debug_info_buffer + off, count);
	if ((off + count) >= info_len) {
		*eof = 1;
		result = info_len - off;
		if (result < 0)
			result = 0;
	} else
		result = count;

	free_pages((unsigned long) debug_info_buffer, 0);
	debug_info_buffer = NULL;
	return result;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
static unsigned long kstat_store = 0;
/* kstat_save
 * Save the contents of the kstat array so that
 * our labours are hidden from vmstat.
 */
static int kstat_save(void)
{
	const int bytes_per_page = PAGE_SIZE - sizeof(unsigned long);
	int kstat_pages = (sizeof(kstat) + bytes_per_page - 1) / bytes_per_page;
	int i, source_offset = 0, bytes;
	unsigned long this_page, last_page = 0;

	kstat_store = this_page = get_zeroed_page(GFP_ATOMIC);
	for (i = 1; i <= kstat_pages; i++) {
		if (!this_page) {
			int k;
			this_page = kstat_store;
			for (k = 1; k < i; k++) {
				unsigned long next_page =
					*((unsigned long *) this_page +
					  ((PAGE_SIZE/sizeof(unsigned long)) -1)
					 );
				free_pages(this_page, 0);
				this_page = next_page;
			}
			kstat_store = 0;
			return -ENOMEM;
		}
		bytes = ((sizeof(kstat) - source_offset) >= bytes_per_page) ?
			bytes_per_page : (sizeof(kstat) - source_offset);
		memcpy((char *) this_page, ((char *) &kstat) + source_offset,
				bytes);
		if (i < kstat_pages) {
			last_page = this_page;
			this_page = get_zeroed_page(GFP_ATOMIC);
			*((unsigned long *) last_page +
			  ((PAGE_SIZE / sizeof(unsigned long)) -1)) =
				this_page;
			source_offset += bytes_per_page;
		}
	}
	return 0;
}

/* kstat_restore
 * Restore a previously saved kstat array
 */
static void kstat_restore(void)
{
	const int bytes_per_page = PAGE_SIZE - sizeof(unsigned long);
	int kstat_pages = (sizeof(kstat) + bytes_per_page - 1) / bytes_per_page;
	int i, source_offset = 0, bytes;
	unsigned long this_page = kstat_store, next_page;

	if (!kstat_store)
		return;

	for (i = 1; i <= kstat_pages; i++) {
		bytes =	((sizeof(kstat) - source_offset) >= bytes_per_page) ?
			bytes_per_page : (sizeof(kstat) - source_offset);
		memcpy(((char *) &kstat) + source_offset,
				(char *) this_page, bytes);
		next_page = *((unsigned long *) this_page + 
				((PAGE_SIZE / sizeof(unsigned long)) - 1));
		source_offset += bytes_per_page;
		free_pages(this_page, 0);
		this_page = next_page;
	}
	kstat_store = 0;
}
#else
#define kstat_save()
#define kstat_restore()
#endif

#ifdef CONFIG_SOFTWARE_SUSPEND_VARIATION_ANALYSIS
/*
 * Checksumming
 */

unsigned long suspend_page_checksum(struct page * page)
{
	unsigned long * virt;
	int i;
	unsigned long value = 0;

	virt = (unsigned long *) kmap(page);
	for (i = 0; i < (PAGE_SIZE / sizeof(unsigned long)); i++)
		value += *(virt + i);
	kunmap(page);
	return value;
}

extern void get_first_pbe(struct pbe * pbe, struct pagedir * pagedir);
extern void get_next_pbe(struct pbe * pbe);

void suspend_calculate_checksums(void)
{
	struct pbe pbe;
	int i = 0, page_index = 0, whichpagedir = 1;
	unsigned long * current_checksum_page = first_checksum_page;
	
	if (!first_checksum_page) {
		prepare_status(1, 0, "Unable to checksum at this point.");
		return;
	}

	prepare_status(1, 0, "Calculating checksums... ");
	
	get_first_pbe(&pbe, &pagedir1);
	
	do {
		*(current_checksum_page + page_index) =
			suspend_page_checksum(pbe.origaddress);
		i++;
		page_index++;
		if (page_index == CHECKSUMS_PER_PAGE) {
			page_index = 0;
			current_checksum_page = (unsigned long *)
				NEXT_CHECKSUM_PAGE(current_checksum_page);
		}
		if (whichpagedir == 1) {
			if (pagedir1.pageset_size == i) {
				if (software_suspend_state & 
					 SOFTWARE_SUSPEND_PAGESET2_NOT_LOADED)
					goto out;
				get_first_pbe(&pbe, &pagedir2);
				whichpagedir = 2;
				i = 0;
			}
		} else {
			if (pagedir2.pageset_size == i)
				goto out;
		}
		get_next_pbe(&pbe);
	} while(1);
	
out:
	prepare_status(1, 0, "Checksums done.");
}

int suspend_page_in_pool(struct page * page);

void suspend_check_checksums(void)
{
	struct pbe pbe;
	int i = 0, page_index = 0, whichpagedir = 1;
	unsigned long * current_checksum_page = first_checksum_page;
	unsigned long sum_now;
	struct reload_data * next_reload_data = first_reload_data;

	if (!first_checksum_page) {
		prepare_status(1, 0, "Unable to checksum at this point.");
		return;
	}

	prepare_status(1, 0, "Checking checksums... ");
	
	get_first_pbe(&pbe, &pagedir1);
	
	do {
		if (PageChecksumIgnore(pbe.origaddress))
			goto skip;
				
		sum_now = suspend_page_checksum(pbe.origaddress);
		if (sum_now != *(current_checksum_page + page_index)) {
			if (next_reload_data) {
				char * virt;
				next_reload_data->pageset = whichpagedir;
				next_reload_data->pagenumber = i;
				next_reload_data->page_address = pbe.origaddress;
				virt = kmap(pbe.origaddress);
				memcpy(next_reload_data->compared_version, 
						virt, PAGE_SIZE);
				kunmap(pbe.origaddress);
				next_reload_data = next_reload_data->next;
			}
		}
skip:
		i++;
		page_index++;
		if (page_index == CHECKSUMS_PER_PAGE) {
			page_index = 0;
			current_checksum_page =	(unsigned long *) 
				NEXT_CHECKSUM_PAGE(current_checksum_page);
		}
		if (whichpagedir == 1) {
			if (pagedir1.pageset_size == i) {
				if (software_suspend_state & 
					 SOFTWARE_SUSPEND_PAGESET2_NOT_LOADED)
					goto out;
				get_first_pbe(&pbe, &pagedir2);
				whichpagedir = 2;
				i = 0;
			}
		} else {
			if (pagedir2.pageset_size == i)
				return;
		}
		get_next_pbe(&pbe);
	
	} while(1);
	
out:
	prepare_status(1, 0, "Differencing done.");
}

/*
 * free_reload_data.
 *
 * Reload data begins on a page boundary.
 */
static void suspend_free_reload_data(void)
{
	struct reload_data * this_data = first_reload_data;
	struct reload_data *prev_reload_data = NULL;
	
	while (this_data) {
		if (this_data->compared_version) {
		  ClearPageNosave(virt_to_page(this_data->compared_version));
		  free_pages((unsigned long) this_data->compared_version, 0);
		}

		if (this_data->base_version) {
			ClearPageNosave(virt_to_page(this_data->base_version));
			free_pages((unsigned long) this_data->base_version, 0);
		}

		if (!(((unsigned long) this_data) & PAGE_MASK))
			prev_reload_data->next = this_data->next;
		else
			prev_reload_data = this_data;
		this_data = this_data->next;
	}

	this_data = first_reload_data;
	while (this_data) {
		prev_reload_data = this_data;
		this_data = this_data->next;
		free_pages((unsigned long) prev_reload_data, 0);
	}

	first_reload_data = last_reload_data = NULL;

}

static void suspend_free_checksum_pages(void)
{
	unsigned long * next_checksum_page;

	while(first_checksum_page) {
		next_checksum_page =
		  (unsigned long *) NEXT_CHECKSUM_PAGE(first_checksum_page);
		free_pages((unsigned long) first_checksum_page, 0);
		first_checksum_page = next_checksum_page;
	}
	last_checksum_page = NULL;
	checksum_pages = 0;
	suspend_store_free_mem(SUSPEND_FREE_CHECKSUM_PAGES, 1);
}

static int page_rangepage(char * seeking, unsigned long * rangepages_list)
{
	int i;
	
	for (i = 1; i <= num_range_pages; i++)
		if (rangepages_list[i] == (unsigned long) seeking)
			return 1;

	return 0;
}

#define PRINTABLE(a) (((a) < 32 || (a) > 122) ? '.' : (a))
static void local_print_location(
		unsigned char * real,
		unsigned char * original,
		unsigned char * resumetime,
		unsigned long * rangepages_list)
{
	int i;

	for (i = 0; i < 8; i++)
		if (*(original + i) != *(resumetime + i))
			break;
	if (i == 8)
		return;
	
	printlog(SUSPEND_INTEGRITY, SUSPEND_HIGH, "%p: ", real);

	for (i = 0; i < 8; i++)
		printlog(SUSPEND_INTEGRITY, SUSPEND_HIGH,
			"%2x ", *(original + i));
	printlog(SUSPEND_INTEGRITY, SUSPEND_HIGH, "    ");
	for (i = 0; i < 8; i++)
		printlog(SUSPEND_INTEGRITY, SUSPEND_HIGH,
			"%c", PRINTABLE(*(original + i)));
	printlog(SUSPEND_INTEGRITY, SUSPEND_HIGH, "    ");
	
	for (i = 0; i < 8; i++)
		printlog(SUSPEND_INTEGRITY, SUSPEND_HIGH,
			"%2x ", *(resumetime + i));
	printlog(SUSPEND_INTEGRITY, SUSPEND_HIGH, "    ");
	for (i = 0; i < 8; i++)
		printlog(SUSPEND_INTEGRITY, SUSPEND_HIGH,
			"%c", PRINTABLE(*(resumetime + i)));
	printlog(SUSPEND_INTEGRITY, SUSPEND_HIGH, "\n");
	if (suspend_page_in_pool(virt_to_page(real)))
		printlog(SUSPEND_INTEGRITY, SUSPEND_HIGH,
			"[In memory pool]");
	if (PageNosave(virt_to_page(real)))
		printlog(SUSPEND_INTEGRITY, SUSPEND_HIGH,
			"[NoSave]");
	if (page_rangepage(real, rangepages_list))
		printlog(SUSPEND_INTEGRITY, SUSPEND_HIGH,
			"[RangePage]");
#ifdef CONFIG_KDB
	{
	int did_print = 0;
	
	for (i = 0; i < 8; i++) {
		static const char *last_sym = NULL;
		if (*(original + i) != *(resumetime + i)) {
			kdb_symtab_t symtab;

			kdbnearsym((unsigned long) real + i,
					&symtab);

			if ((!symtab.sym_name) ||
					(symtab.sym_name == last_sym))
				continue;

			last_sym = symtab.sym_name;

			printlog(SUSPEND_INTEGRITY, SUSPEND_LOW,
				"%p = %s\n",
				symtab.sym_start,
				symtab.sym_name);
			did_print = 1;
		}
	}

	if (did_print)
		printlog(SUSPEND_INTEGRITY, SUSPEND_LOW,
			"\n");
	}
#endif
}


static void suspend_print_differences(void)
{
	struct reload_data * this_data = first_reload_data;
	int i;
	unsigned long * rangepages = get_rangepages_list();

	if (!rangepages)
		return;

	while (this_data) {
		if (this_data->pageset && 
		    this_data->pagenumber) {
			printlog(SUSPEND_INTEGRITY, SUSPEND_MEDIUM,
				"Pagedir %d. Page %d. Address %p."
				" Base %p. Copy %p.\n",
				this_data->pageset,
				this_data->pagenumber,
				page_address(this_data->page_address),
				this_data->base_version,
				this_data->compared_version);
			for (i= 0; i < (PAGE_SIZE / 8); i++) {
				local_print_location(
					page_address(this_data->page_address) + i * 8,
					this_data->base_version + i * 8,
					this_data->compared_version + i * 8,
					rangepages);
				check_shift_keys(0, NULL);
			}
			check_shift_keys(1, NULL);
		} else
			return;
		this_data = this_data->next;
	}

	put_rangepages_list(rangepages);
}

static int suspend_allocate_reload_data(int pages)
{
	struct reload_data * this_data;
	unsigned long data_start;
	int i;

	for (i = 1; i <= pages; i++) {
		data_start = get_grabbed_pages(0);

		if (!data_start)
			return -ENOMEM;

		SetPageChecksumIgnore(virt_to_page(data_start));
		this_data = (struct reload_data *) data_start; 

		while (data_start == 
		  ((((unsigned long) (this_data + 1)) - 1) & PAGE_MASK)) {
			struct page * page;
			unsigned long virt;

			virt = get_grabbed_pages(0);
			if (!virt) {
				printk("Couldn't get a page in which to store "
					"a changed page.\n");
				return -ENOMEM;
			}
			page = virt_to_page(virt);

			this_data->compared_version = (char *) virt;
			SetPageNosave(page);
			SetPageChecksumIgnore(page);
			
			virt = get_grabbed_pages(0);
			if (!virt) {
				printk("Couldn't get a page in which to store "
					"a baseline page.\n");
				return -ENOMEM;
			}
			page = virt_to_page(virt);

			this_data->base_version = (char *) virt;
			SetPageNosave(page);
			SetPageChecksumIgnore(page);

			if (last_reload_data)
				last_reload_data->next = this_data;
			else
				first_reload_data = this_data;
			
			last_reload_data = this_data;

			this_data++;
		}

		check_shift_keys(0, NULL);
	}

	return 0;
}

void suspend_allocate_checksum_pages(void)
{
	int pages_required =
		(pageset1_size + pageset2_size) / CHECKSUMS_PER_PAGE;
	unsigned long this_page;

	while (checksum_pages <= pages_required) {
		this_page = get_grabbed_pages(0);
		if (!first_checksum_page)
			first_checksum_page = 
				(unsigned long *) this_page;
		else
			NEXT_CHECKSUM_PAGE(last_checksum_page) = this_page;
		
		last_checksum_page = (unsigned long *) this_page;
		SetPageChecksumIgnore(virt_to_page(this_page));
		checksum_pages++;
	}
	suspend_store_free_mem(SUSPEND_FREE_CHECKSUM_PAGES, 0);
}
#endif

#define SUSPEND_C
#include <asm/suspend.h>

/*
 * software_suspend_pending
 * Functionality   : First level of code for software suspend invocations.
 *                   Stores and restores load averages (to avoid a spike),
 *                   allocates bitmaps, freezes processes and eats memory
 *                   as required before suspending drivers and invoking
 *                   the 'low level' code to save the state to disk.
 *                   By the time we return from do_suspend2_lowlevel, we
 *                   have either failed to save the image or successfully
 *                   suspended and reloaded the image. The difference can
 *                   be discerned by checking SUSPEND_ABORTED.
 * Called From     : 
 */

static int get_suspend_debug_info(void);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
extern inline int notify_resume(void);
#if defined(CONFIG_MAGIC_SYSRQ) && defined(CONFIG_VT)
extern int sysrq_pressed;
#endif
#endif

static int can_suspend(void)
{
	if (software_suspend_state & 2) {
		printk(name_suspend "Software suspend is already running.\n");
		return 0;
	}

	if (software_suspend_state & 1) {
		printk(name_suspend "Software suspend is disabled.\n"
			"This may be because you haven't put something along the "
			"lines of\n\nresume2=swap:/dev/hda1\n\n"
			"in lilo.conf or equivalent. (Where /dev/hda1 is your "
			"swap partition).\n");
		SET_RESULT_STATE(SUSPEND_ABORTED);
		return 0;
	}

	BUG_ON(in_interrupt());
	
	return 1;
}

static int allocate_bitmaps(void)
{
	printlog(SUSPEND_MEMORY, SUSPEND_VERBOSE, "Allocating in_use_map\n");
	if (allocatemap(&in_use_map, 1))
		return 1;
	
	PRINTFREEMEM("after allocating in_use_map");
	suspend_store_free_mem(3, 0);
	
	if (allocatemap(&pageset2_map, 1))
		return 1;
	
	PRINTFREEMEM("after allocating pageset2 map");
	suspend_store_free_mem(4, 0);
	
#ifdef CONFIG_SOFTWARE_SUSPEND_VARIATION_ANALYSIS
	/* Allocate even if not used to save having extra tests elsewhere. */
	if (allocatemap(&checksum_map, 0))
		return 1;
	PRINTFREEMEM("after allocating checksum map");
	suspend_store_free_mem(5, 0);
#endif

	return 0;
}

#if defined(CONFIG_SOFTWARE_SUSPEND_DEBUG) && LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
void show_pcp_lists(void)
{
	int cpu, temperature;
	struct zone *zone;

	for_each_zone(zone) {
		printk("%s per-cpu:", zone->name);

		if (!zone->present_pages) {
			printk(" empty\n");
			continue;
		} else
			printk("\n");

		for (cpu = 0; cpu < NR_CPUS; ++cpu) {
			struct per_cpu_pageset *pageset;

			if (!cpu_possible(cpu))
				continue;

			pageset = zone->pageset + cpu;

			for (temperature = 0; temperature < 2; temperature++)
				printk("cpu %d %s: low %d, high %d, batch %d, count %d.\n",
					cpu,
					temperature ? "cold" : "hot",
					pageset->pcp[temperature].low,
					pageset->pcp[temperature].high,
					pageset->pcp[temperature].batch,
					pageset->pcp[temperature].count);
		}
	}
}
#endif

void software_suspend_pending(void)
{
	unsigned long avenrun_save[3];
	int i;
	mm_segment_t	oldfs;

	if (!can_suspend())
		return;
	
        /* Suspend always runs on processor 0 */ 
	ensure_on_processor_zero();

	display_nosave_pages();

#ifdef CONFIG_SOFTWARE_SUSPEND_KEEP_IMAGE
	if (TEST_RESULT_STATE(SUSPEND_KEPT_IMAGE)) {
		if (TEST_ACTION_STATE(SUSPEND_KEEP_IMAGE)) {
			printk("Image already stored:"
				" powering down immediately.");
			suspend_power_down();
			return;	/* It shouldn't, but just in case */
		} else {
			printk("Invalidating previous image.\n");
			active_writer->ops.writer.invalidate_image();
		}
	}
#endif

	printk(name_suspend "Initiating a software_suspend cycle.\n");
	software_suspend_state |= SOFTWARE_SUSPEND_RUNNING;
	oldfs = get_fs(); set_fs(KERNEL_DS);

	suspend_result = 0;
	max_ranges_used = 0;
	nr_suspends++;
	now_resuming = 0;
	
	suspend_io_time[0][0] = suspend_io_time[0][1] = suspend_io_time[1][0] =
		suspend_io_time[1][1] = 0;

	PRINTFREEMEM("at start of software_suspend_pending");
	suspend_store_free_mem(SUSPEND_FREE_BASE, 0);

	/*
	 * Running suspend makes for a very high load average. I'm told that
	 * sendmail and crond check the load average, so in order for them
	 * not to be unnecessarily affected by the operation of suspend, we
	 * store the avenrun values prior to suspending and restore them
	 * at the end of the resume cycle. Thus, the operation of suspend
	 * should be invisible to them. Thanks to Marcus Gaugusch and Bernard
	 * Blackham for noticing the problem and suggesting the solution.
	 */
	
	for (i = 0; i < 3; i++)
		avenrun_save[i] = avenrun[i];

	/* Suspend console switch (if necessary) */
	if (pm_prepare_console())
		printk(name_suspend "Can't allocate a console... proceeding\n");

	PRINTFREEMEM("after preparing suspend_console");
	suspend_store_free_mem(SUSPEND_FREE_CONSOLE_ALLOC, 0);

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
	disable_pcp_lists();

	PRINTFREEMEM("after draining local pages");
	suspend_store_free_mem(SUSPEND_FREE_DRAIN_PCP, 0);

	if (TEST_DEBUG_STATE(SUSPEND_FREEZER))
		show_pcp_lists();
#endif

	if (allocate_bitmaps())
		goto out;
	
	PRINTFREEMEM("after allocating bitmaps");

#ifdef CONFIG_SOFTWARE_SUSPEND_VARIATION_ANALYSIS
	if (CHECKMASK(SUSPEND_INTEGRITY)) {
		suspend_allocate_reload_data(1);
		PRINTFREEMEM("after allocating reload data");
		suspend_store_free_mem(SUSPEND_FREE_RELOAD_PAGES, 0);
	}
#endif

	display_nosave_pages();

	set_chain_names(&pagedir1);
	set_chain_names(&pagedir2);

	if (initialise_suspend_plugins())
		goto out;

	PRINTFREEMEM("after initialising plugins");
	suspend_store_free_mem(SUSPEND_FREE_INIT_PLUGINS, 0);

	/* Free up memory if necessary */
	printlog(SUSPEND_ANY_SECTION, SUSPEND_VERBOSE, "Preparing image.\n");
	PRINTPREEMPTCOUNT("Before preparing image.");
	if (prepare_image() || TEST_RESULT_STATE(SUSPEND_ABORTED))
		goto out;

	PRINTPREEMPTCOUNT("After prepare_image.");
	PRINTFREEMEM("after preparing image");

	if (TEST_ACTION_STATE(SUSPEND_FREEZER_TEST))
		goto out;

	/* We don't want suspend to show in the kernel statistics -
	 * it should be transparent to userspace */
	kstat_save();

	PRINTFREEMEM("after saving kernel stats");
	suspend_store_free_mem(SUSPEND_FREE_KSTAT, 0);

	display_nosave_pages();

	if (!TEST_RESULT_STATE(SUSPEND_ABORTED)) {
		prepare_status(1, 0, "Starting to save the image..");
		save_image();
		PRINTPREEMPTCOUNT("After saving the image.");
	}
	
out:
	PRINTFREEMEM("at 'out'");

	i = get_suspend_debug_info();

	PRINTPREEMPTCOUNT("After storing debug info.");
	suspend_store_free_mem(SUSPEND_FREE_DEBUG_INFO, 0);

	software_suspend_state &= ~SOFTWARE_SUSPEND_USE_MEMORY_POOL;

#ifdef CONFIG_SOFTWARE_SUSPEND_VARIATION_ANALYSIS
	suspend_free_checksum_pages();
	PRINTFREEMEM("after freeing checksum pages");
	suspend_store_free_mem(SUSPEND_FREE_CHECKSUM_PAGES, 1);

	if (CHECKMASK(SUSPEND_INTEGRITY)) {
		suspend_free_reload_data();
		PRINTFREEMEM("after freeing reload data");
		suspend_store_free_mem(SUSPEND_FREE_RELOAD_PAGES, 1);
	}
#endif

	STORAGE_UNSUSPEND

	free_pagedir(&pagedir2);
	PRINTFREEMEM("after freeing pagedir 1");
	free_pagedir(&pagedir1);
	PRINTFREEMEM("after freeing pagedir 2");
	
	empty_suspend_memory_pool();
	PRINTFREEMEM("after freeing memory pool");
	suspend_store_free_mem(SUSPEND_FREE_MEM_POOL, 1);

	free_ranges();
	suspend_store_free_mem(SUSPEND_FREE_RANGE_PAGES, 1);
	PRINTFREEMEM("after freeing ranges");

	freemap(&checksum_map);
	PRINTFREEMEM("after freeing checksum map");
	suspend_store_free_mem(SUSPEND_FREE_CHECKSUM_MAP, 1);

	freemap(&pageset2_map);
	PRINTFREEMEM("after freeing pageset2 map");
	suspend_store_free_mem(SUSPEND_FREE_PS2_MAP, 1);

	freemap(&in_use_map);
	PRINTFREEMEM("after freeing inuse map");
	suspend_store_free_mem(SUSPEND_FREE_IN_USE_MAP, 1);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#if defined(CONFIG_MAGIC_SYSRQ) && defined(CONFIG_VT)
	sysrq_pressed = 0;
#endif
#endif
	/* Restore stats before we restart processes */
	for (i = 0; i < 3; i++)
		avenrun[i] = avenrun_save[i];
	
	if (debug_info_buffer) {
		/* Printk can only handle 1023 bytes, including
		 * its level mangling. */
		for (i = 0; i < 3; i++)
			printk("%s", debug_info_buffer + (1023 * i));
		free_pages((unsigned long) debug_info_buffer, 0);
		debug_info_buffer = NULL;
	}
	PRINTFREEMEM("after freeing debug info buffer");
	suspend_store_free_mem(SUSPEND_FREE_DEBUG_INFO, 1);
	
	cleanup_suspend_plugins();

	PRINTFREEMEM("after cleaning up suspend plugins");
	suspend_store_free_mem(SUSPEND_FREE_INIT_PLUGINS, 1);
	
	kstat_restore();

	PRINTFREEMEM("after restoring kstats");
	suspend_store_free_mem(SUSPEND_FREE_KSTAT, 1);
	
	display_nosave_pages();

	thaw_processes();

	PRINTFREEMEM("after thawing processes");
	suspend_store_free_mem(SUSPEND_FREE_FREEZER, 1);

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
	enable_pcp_lists();
	suspend_store_free_mem(SUSPEND_FREE_DRAIN_PCP, 1);

	if (TEST_DEBUG_STATE(SUSPEND_FREEZER))
		show_pcp_lists();
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	if (notify_resume())
		printk(KERN_EMERG "Failed to notify resume chain.\n");
#endif

	MDELAY(1000);
	
	check_shift_keys(1, "About to restore original console.");
	suspend_store_free_mem(SUSPEND_FREE_FREEZER, 1);

	pm_restore_console();
	suspend_store_free_mem(SUSPEND_FREE_CONSOLE_ALLOC, 1);

	suspend_store_free_mem(SUSPEND_FREE_BASE, 1);
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	display_free_mem();
#endif

	software_suspend_state &= ~SOFTWARE_SUSPEND_RUNNING;
	PRINTFREEMEM("at end of software_suspend_pending");
#ifdef CONFIG_PREEMPT
	PRINTPREEMPTCOUNT("Exiting with preempt count");
#endif
	set_fs(oldfs);
}

int suspend_register_plugin(struct suspend_plugin_ops * plugin)
{
	if (!(num_filters + num_writers)) {
		INIT_LIST_HEAD(&suspend_filters);
		INIT_LIST_HEAD(&suspend_writers);
		INIT_LIST_HEAD(&suspend_plugins);
	}

	switch (plugin->type) {
		case FILTER_PLUGIN:
			list_add_tail(&plugin->ops.filter.filter_list,
					&suspend_filters);
			num_filters++;
			break;

		case WRITER_PLUGIN:
			list_add_tail(&plugin->ops.writer.writer_list,
					&suspend_writers);
			num_writers++;
			break;

		default:
			printk("Hmmm. Plugin '%s' has an invalid type."
				" It has been ignored.\n", plugin->name);
			return -EINVAL;
	}
	list_add(&plugin->plugin_list, &suspend_plugins);

	return 0;	
}

int initialise_suspend_plugins(void)
{
	struct list_head *plugin;
	struct suspend_plugin_ops * this_plugin;
	int result;
	
	list_for_each(plugin, &suspend_plugins) {
		this_plugin = list_entry(plugin, struct suspend_plugin_ops,
				plugin_list);
		if (this_plugin->disabled)
			continue;
		printlog(SUSPEND_MEMORY, SUSPEND_MEDIUM,
			"Initialising plugin %s.\n",
			this_plugin->name);
		if (this_plugin->initialise)
			if ((result = this_plugin->initialise()))
				return result;
		PRINTFREEMEM("after initialising plugin");
	}

	return 0;
}

void cleanup_suspend_plugins(void)
{
	struct list_head *plugin;
	struct suspend_plugin_ops * this_plugin;
	
	list_for_each(plugin, &suspend_plugins) {
		this_plugin = list_entry(plugin, struct suspend_plugin_ops, 
				plugin_list);
		if (this_plugin->disabled)
			continue;
		printlog(SUSPEND_MEMORY, SUSPEND_MEDIUM,
			"Cleaning up plugin %s.\n",
			this_plugin->name);
		if (this_plugin->cleanup)
			this_plugin->cleanup();
		PRINTFREEMEM("after cleaning up plugin");
	}
}

struct suspend_plugin_ops * 
get_next_filter(struct suspend_plugin_ops * filter_sought)
{
	struct list_head *filter;
	struct suspend_plugin_ops * last_filter = NULL, *this_filter = NULL;

	list_for_each(filter, &suspend_filters) {
		this_filter = list_entry(filter, struct suspend_plugin_ops,
				ops.filter.filter_list);
		if (this_filter->disabled)
			continue;
		if ((last_filter == filter_sought) || (!filter_sought))
			return this_filter;
		last_filter = this_filter;
	}

	return active_writer;
}

int attempt_to_parse_resume_device(int boot_time)
{
	struct list_head *writer;
	struct suspend_plugin_ops * this_writer;
	int result = 0;

	if (!num_writers) {
		printk(name_suspend "No writers have been registered.\n");
		return 0;
	}
	
	software_suspend_state &= ~SOFTWARE_SUSPEND_RESUME_DEVICE_OK;
	software_suspend_state |= SOFTWARE_SUSPEND_DISABLED;

	if (!resume_file[0])
		return -EINVAL;

	list_for_each(writer, &suspend_writers) {
		this_writer = list_entry(writer, struct suspend_plugin_ops,
				ops.writer.writer_list);

		/* 
		 * Not sure why you'd want to disable a writer, but
		 * we should honour the flag if we're providing it
		 */
		if (this_writer->disabled)
			continue;

		result = this_writer->ops.writer.parse_image_location(
				resume_file, boot_time, (num_writers == 1));

		switch (result) {
			case -EINVAL:
				/* 
				 * For this writer, but not a valid 
				 * configuration
				 */

				printk(name_suspend
					"Not able to successfully parse this "
					"resume device. Suspending disabled.\n");
				return result;

			case 0:
				/*
				 * Not for this writer. Try the next one.
				 */

				break;

			case 1:
				/*
				 * For this writer and valid.
				 */

				active_writer = this_writer;

				/* We may not have any filters compiled in */

				software_suspend_state |=
					SOFTWARE_SUSPEND_RESUME_DEVICE_OK;
				software_suspend_state &= ~3;
				printk(name_suspend "Suspending enabled.\n");
				return 0;
		}
	}
	printk(name_suspend "No matching writer found. Suspending disabled.\n");
	return -EINVAL;
}

/*
 * Called from init kernel_thread.
 * We check if we have an image and if so we try to resume.
 * We also start ksuspendd if configuration looks right.
 */

extern int freeze_processes(int no_progress);
extern int orig_loglevel;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
void software_resume2(void)
{
#define RESUME2_RET
#else
static int __init software_resume2(void)
{
#define RESUME2_RET ret
	int ret = 0;
#endif
	int read_image_result = 0;
	orig_loglevel = console_loglevel;

	/* Suspend always runs on processor 0 */
	ensure_on_processor_zero();

	PRINTPREEMPTCOUNT("at start of software_resume2");

	if (sizeof(swp_entry_t) != sizeof(long)) {
		printk(KERN_WARNING name_suspend
			"The size of swp_entry_t != size of long. "
			"Please report this!\n");
		return RESUME2_RET;
	}
	
	software_suspend_state |= SOFTWARE_SUSPEND_RUNNING;

	if (!resume_file[0])
		printk(KERN_WARNING name_suspend
			"You need to use a resume2= command line parameter to "
			"tell Software Suspend 2 where to look for an image.\n");
	else
		attempt_to_parse_resume_device(1);
	
	suspend_init_proc();

	if (!(software_suspend_state & SOFTWARE_SUSPEND_RESUME_DEVICE_OK)) {

		/* 
		 * Without a usable storage device we can do nothing - 
		 * even if noresume is given
		 */

		if (!num_writers)
			printk(KERN_ALERT name_suspend
				"No writers have been registered. You must "
				"select a method of storing the image when "
				"compiling the kernel.\n");
		else
			printk(KERN_ALERT name_suspend
				"Missing or invalid storage location "
				"(resume2= parameter). Please correct and "
				"rerun lilo (or equivalent) before "
				"suspending.\n");
		software_suspend_state &= ~SOFTWARE_SUSPEND_RUNNING;
		return RESUME2_RET;
	}

	/* We enable the possibility of machine suspend */
	orig_mem_free = nr_free_pages();

	suspend_task = current->pid;

	printk(name_suspend "Checking for image...\n");

	PRINTPREEMPTCOUNT("when about to read primary image");
	
	read_image_result = read_primary_suspend_image(); /* non fatal error ignored */

	if (software_suspend_state & SOFTWARE_SUSPEND_NORESUME_SPECIFIED)
		printk(KERN_WARNING name_suspend "Resuming disabled as requested.\n");

	if (read_image_result) {
		suspend_task = 0;
		software_suspend_state &= ~3;
		console_loglevel = orig_loglevel;
		return RESUME2_RET;
	}

	freeze_processes(1);

	prepare_status(0, 0,
		"Copying original kernel back (no status - sensitive!)...");
	
	PRINTPREEMPTCOUNT("Prior to calling do_suspend2_lowlevel.");
        
	do_suspend2_lowlevel(1);
	BUG();

	return RESUME2_RET;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
late_initcall(software_resume2);
#endif

/*
 * Resume setup: obtain the storage device.
 */

static int __init resume_setup(char *str)
{
	if (str == NULL)
		return 1;
	
	strncpy(resume_file, str, 255);
	return 0;
}

/*
 * Allow the user to set the action parameter from lilo, prior to resuming.
 */
static int __init suspend_act_setup(char *str)
{
	if(str)
		suspend_action=simple_strtol(str,NULL,0);
	suspend_act_used = 1;
	return 0;
}

/*
 * Allow the user to set the debug parameter from lilo, prior to resuming.
 */
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
static int __init suspend_dbg_setup(char *str)
{
	if(str)
		suspend_debug_state=simple_strtol(str,NULL,0);
	suspend_dbg_used = 1;
	return 0;
}

/*
 * Allow the user to set the debug level parameter from lilo, prior to
 * resuming.
 */
static int __init suspend_lvl_setup(char *str)
{
	if(str)
		console_loglevel =
		suspend_default_console_level = 
			simple_strtol(str,NULL,0);
	suspend_lvl_used = 1;
	return 0;
}
#endif

/*
 * Allow the user to specify that we should ignore any image found and
 * invalidate the image if necesssary. This is equivalent to running
 * the task queue and a sync and then turning off the power. The same
 * precautions should be taken: fsck if you're not journalled.
 */
static int __init noresume_setup(char *str)
{
	software_suspend_state |= SOFTWARE_SUSPEND_NORESUME_SPECIFIED;
	/* Message printed later */
	return 0;
}

__setup("resume2=", resume_setup);
__setup("suspend_act=", suspend_act_setup);
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
__setup("suspend_dbg=", suspend_dbg_setup);
__setup("suspend_lvl=", suspend_lvl_setup);
#endif
__setup("noresume2", noresume_setup);

EXPORT_SYMBOL(software_suspend_pending);
