/*
 * kernel/power/ui.c
 *
 * Copyright (C) 1998-2001 Gabor Kuti <seasons@fornax.hu>
 * Copyright (C) 1998,2001,2002 Pavel Machek <pavel@suse.cz>
 * Copyright (C) 2002-2003 Florent Chabaud <fchabaud@free.fr>
 * Copyright (C) 2002-2004 Nigel Cunningham <ncunningham@users.sourceforge.net>
 *
 * This file is released under the GPLv2.
 *
 * Routines for Software Suspend's user interface.
 * 
 * The user interface includes support for both a 'nice display' and
 * a bootsplash screen (bootsplash.org), and for run-time debugging.
 *
 * The 'nice display' is text based and implements a progress bar and
 * (optional) textual progress, as well as an overall description of
 * the current action and the display of a header and the code version.
 *
 * The bootsplash support uses calls to bootsplash's proc interface
 * to set the progress bar value. These calls replace the display of
 * the text based progress bar only; if the bootsplash is in verbose
 * mode, the header, version and description still show. In silent
 * mode, the calls are still made but the text is not seen.
 *
 * Note that this implies a bootsplash picture version of 3 or later.
 * Earlier ones require a separate bootsplash option patch.
 *
 * As well as displaying status information, the user has some control
 * over the software while suspending. Hooks in drivers/char/console.c
 * and drivers/char/serial.c allow a user on the keyboard or serial
 * console to...
 *
 * 					 	  Key
 * Toggle rebooting			 	   R
 * Toggle logging all output		 	   L
 * Toggle pausing between major steps (1)	   P
 * Toggle pausing at minor steps		   S
 * Switch log levels (2)			  0-7
 * Cancel the suspend (3)			 Escape
 * Continue when paused				 Space
 *
 * (1) Pausing only occurs when the log level is > 1.
 * (2) When debugging is not compiled in, only levels 0 & 1 work.
 * (3) Can be disabled using /proc/suspend/enable_escape.
 *
 * (The following assumes debugging is compiled in)...
 * 
 * Fundamental to the debugging code is the idea that each debug 
 * message which suspend can display has both a section of code to
 * which it belongs and a level of verbosity. When suspend is running,
 * it only displays a message if debugging for the section has been
 * enabled prior to beginning the cycle and the current loglevel is
 * greater than or equal to the level of the message.
 * 
 * Prior to starting a suspend cycle, a user can select which sections
 * should display debugging information by setting
 * /proc/suspend/debug_sections. This is a bit vector (values in
 * include/linux/suspend-debug.h). The initial log level can be also
 * be set using /proc/suspend/default_console_level.
 * The debug sections and level can be changed prior to resuming using
 * the kernel command line parameters suspend_dbg and suspend_lvl.
 * 
 * In addition to the above ability to control whether messages are
 * displayed, messages can be displayed in two ways. The printlog call
 * displays a message using printk, with the result that it is also
 * logged to syslog in the normal way.
 *
 * A call to printnolog usually gets the text to the screen using
 * vt_console_print and is thus not logged. This is the preferred
 * means of displaying highlevel debugging information, because it
 * reduces clutter in the logs. (You can easily end up with 1/5645.^M
 * 2/5645.^M 3/5645.^M... otherwise).
 * If loggging of this output is wanted, the log all output toggle
 * can be activated and printk will be used instead of
 * vt_console_print.
 */
#define SWSUSP_CONSOLE_C

#define __KERNEL_SYSCALLS__

#include <linux/suspend-common.h>
#include <linux/console.h>
 
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
#include <linux/syscalls.h>
#else
extern asmlinkage long sys_mount(char * dev_name, char * dir_name, char * type,
	unsigned long flags, void * data);
extern asmlinkage long sys_umount(char * name, int flags);
extern asmlinkage long sys_mkdir(const char *name, int mode);
#endif

static DECLARE_WAIT_QUEUE_HEAD(suspend_wait_for_key);

#ifdef DEFAULT_SUSPEND_CONSOLE
static int barwidth = 100, barposn = -1, newbarposn = 0;
static int orig_kmsg;
extern int suspend_default_console_level;
#endif

#ifdef CONFIG_VT
static int orig_fgconsole;
#endif
int orig_loglevel = 0;

static char print_buf[1024];	/* Same as printk - should be safe */

extern void hide_cursor(int currcons);

#if CONFIG_VT_CONSOLE

/*
 * We want the number of lines & columns for the suspend console, not the 
 * current console
 */
#undef video_num_columns
#define video_num_columns	(vc_cons[fg_console].d->vc_cols)
#undef video_num_lines
#define video_num_lines		(vc_cons[fg_console].d->vc_rows)

extern void reset_terminal(int currcons, int do_clear);

/* Forward declaration */
void force_console_redraw(void);

int suspend_console_fd = -1;
struct termios termios;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
extern asmlinkage ssize_t sys_write(unsigned int fd, const char __user * buf, 
	size_t count);
#define write sys_write
#endif

#ifdef CONFIG_DEVFS_FS
static int mounted_devfs = 0;
#endif

#define cond_console_print(chars, count) \
	if (suspend_console_fd > -1) { \
		write(suspend_console_fd, chars, count); \
		acquire_console_sem(); \
		hide_cursor(fg_console); \
		poke_blanked_console(); \
		release_console_sem(); \
	}

static void move_cursor_to(unsigned char * xy)
{
	char buf[10];

	int len = suspend_snprintf(buf, 10, "\233%d;%dH", xy[1], xy[0]);

	cond_console_print(buf, len);
}

static void clear_display(void)
{
	char buf[4] = "\2332J";
	unsigned char home[2] = { 0, 0 };
	
	cond_console_print(buf, 3);
	move_cursor_to(home);
	//force_console_redraw();
}

/* Your bootsplash progress bar may have a width of (eg) 1024 pixels. That
 * doesn't necessarily mean you want the bar updated 1024 times when writing
 * the image */
static int bar_granularity_limit = 0;

/* We remember the last header that was (or could have been) displayed for
 * use during log level switches */
static char lastheader[512];
static int lastheader_message_len = 0;
#endif

/* Remember the last loglevel so we can tell when it changes */
static int lastloglevel = -1;
static int orig_default_message_loglevel;

/* ------------------  Splash screen defines  -------------------------- */

#if defined(CONFIG_PROC_FS) && \
	(defined(CONFIG_BOOTSPLASH) || defined(CONFIG_FBCON_SPLASHSCREEN))
extern struct display fb_display[MAX_NR_CONSOLES];
extern int splash_verbose(void);

/* splash_is_on
 * 
 * Description: Determine whether a VT has a splash screen on.
 * Arguments:	int consolenr. The VT number of a console to check.
 * Returns:	Boolean indicating whether the splash screen for
 *		that console is on right now.
 */
static int splash_is_on(int consolenr)
{
	struct splash_data *info = get_splash_data(consolenr);

	if (info)
		return ((info->splash_state & 1) == 1);
	return 0;
}

/* splash_write_proc.
 *
 * Write to Bootsplash's proc entry. We need this to work when /proc
 * hasn't been mounted yet and / can't be mounted. In addition, we
 * want it to work despite the fact that bootsplash (under 2.4 at least)
 * removes its proc entry when it shouldn't.  We therefore use
 * our proc.c find_proc_dir_entry routine to get the location of the
 * write routine once (boot time & at start of each resume), and keep it.
 */

extern struct proc_dir_entry * find_proc_dir_entry(const char *name,
	struct proc_dir_entry *parent);

static void splash_write_proc(const char *buffer, unsigned long count)
{
	static write_proc_t * write_routine;
	struct proc_dir_entry * proc_entry;

	if (in_interrupt())
		return;

	if (unlikely(!write_routine)) {
		proc_entry = find_proc_dir_entry("splash", &proc_root);
		if (proc_entry)
			write_routine = proc_entry->write_proc;
	}
        
	if (write_routine)
                write_routine(NULL, buffer, count, NULL);
}

/* fb_splash-set_progress
 *
 * Description:	Set the progress bar position for a splash screen.
 * Arguments:	int consolenr. The VT number of a console to use.
 * 		unsigned long value, unsigned long maximum:
 * 		The proportion (value/maximum) of the bar to fill.
 */

static int fb_splash_set_progress(int consolenr, unsigned long value,
	unsigned long maximum)
{
	char procstring[15];
	int length, bitshift = generic_fls(maximum) - 16;
	static unsigned long lastvalue = 0;
	unsigned long thisvalue;

	BUG_ON(consolenr >= MAX_NR_CONSOLES);
	
	if (in_interrupt())
		return 0;

	if (value > maximum)
		value = maximum;

	/* Avoid math problems - we can't do 64 bit math here
	 * (and don't need it - anyone got screen resolution
	 * of 65536 pixels or more?) */
	if (bitshift > 0) {
		maximum = maximum >> bitshift;
		value = value >> bitshift;
	}

	thisvalue = value * 65534 / maximum;

	length = sprintf(procstring, "show %lu", thisvalue);

	splash_write_proc(procstring, length);
	
	/* Ensure redraw when the progress bar goes to a lower value */
	if (thisvalue < lastvalue)
		force_console_redraw();

	lastvalue = thisvalue;
	
	return 0;
}
#else
#define splash_is_on(consolenr) (0)
#define fb_splash_set_progress(...) do { } while(0)
#define splash_write_proc(...) do { } while(0)
#define splash_verbose()
#endif

/* abort_suspend
 *
 * Description: Begin to abort a cycle. If this wasn't at the user's request
 * 		(and we're displaying output), tell the user why and wait for
 * 		them to acknowledge the message.
 * Arguments:	A parameterised string (imagine this is printk) to display,
 *	 	telling the user why we're aborting.
 */

void abort_suspend(const char *fmt, ...)
{
	va_list args;
	int printed_len = 0;

	if (!TEST_RESULT_STATE(SUSPEND_ABORTED)) {
		if ((!TEST_RESULT_STATE(SUSPEND_ABORT_REQUESTED)) 
				&& (!NO_OUTPUT_OR_PAUSING)) {
			int locked = (spin_is_locked(&suspend_irq_lock));

			if (locked)
				spin_unlock_irqrestore(&suspend_irq_lock, 
						suspendirqflags);
			va_start(args, fmt);
			printed_len = vsnprintf(print_buf, 
					sizeof(print_buf), fmt, args);
			va_end(args);
			printed_len = sprintf(print_buf + printed_len,
					" (Press SPACE to continue)");
			prepare_status(1, 1, print_buf);
			/* 
			 * Make sure message seen - wait for shift to be
			 * released if being pressed 
			 */
			interruptible_sleep_on(&suspend_wait_for_key);

			if (locked)
				spin_lock_irqsave(&suspend_irq_lock, 
						suspendirqflags);
		}
		/* Turn on aborting flag */
		SET_RESULT_STATE(SUSPEND_ABORTED);
	}
}

/* handle_loglevel_change
 *
 * Description:	Update the display when the user changes the log level.
 * Returns:	Boolean indicating whether the level was changed.
 */

static int handle_loglevel_change(void)
{
	static int recursive = 0;
	
	if ((console_loglevel == lastloglevel) || recursive)
		return 0;
	
	if (lastloglevel == -1) {
		lastloglevel = console_loglevel;
		return 0;
	}

	recursive = 1;

#ifdef CONFIG_VT
	/* Calculate progress bar width. Note that whether the
	 * splash screen is on might have changed (this might be
	 * the first call in a new cycle), so we can't take it
	 * for granted that the width should be the same as
	 * last time we came in here */
	if (splash_is_on(fg_console)) {
		/* proc interface ensures bar_granularity_limit >= 0 */
		if (bar_granularity_limit)
			barwidth = bar_granularity_limit;
		else
			barwidth = 100;
	} else 
		barwidth = 
			(video_num_columns - 2 * (video_num_columns / 4) - 2);

	/* Only reset the display if we're switching between nice display
	 * and displaying debugging output */
	
	if (console_loglevel > 1) {
		if (lastloglevel < 2) {
			clear_display();
			if (splash_is_on(fg_console))
				splash_write_proc("verbose\n", 9);
		}
#endif
		printnolog(0, console_loglevel, 0,
			"Switched to console loglevel %d.\n", 
			console_loglevel);
#ifdef CONFIG_VT
	} else if (lastloglevel > 1) {
		clear_display();
		if (splash_is_on(fg_console))
			splash_write_proc("silent\n", 8);
	}
	
	/* Get the nice display or last action [re]drawn */
	prepare_status(0, 0, NULL);
#endif
	
	lastloglevel = console_loglevel;

	recursive = 0;
	
	return 1;
}

#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
/* printnolog.
 *
 * Description:	This function is intended to do the same job as printk, but
 * 		without normally logging what is printed. The point is to be
 * 		able to get debugging info on screen without filling the logs
 * 		with "1/534. ^M 2/534^M. 3/534^M"
 *
 * 		It may be called from an interrupt context - can't sleep!
 *
 * Arguments:	int mask: The debugging section(s) this message belongs to.
 * 		int level: The level of verbosity of this message.
 * 		int restartline: Whether to output a \r or \n with this line
 * 			(\n if we're logging all output).
 * 		const char *fmt, ...: Message to be displayed a la printk.
 */
//void printnolog(int mask, int level, int restartline, const char *fmt, ...)
void __printnolog(int restartline, const char *fmt, ...)
{
#if defined(DEFAULT_SUSPEND_CONSOLE)
	va_list args;
	int printed_len = 0;

	/* Don't do this if not printing anything - print[no]log get called
	 * before we prepare the console at resume time */
	handle_loglevel_change();

	if ((restartline) && (!TEST_ACTION_STATE(SUSPEND_LOGALL)))
		printed_len = vsnprintf(print_buf,
				sizeof(print_buf), "\r", NULL);

	va_start(args, fmt);
	printed_len += vsnprintf(print_buf + printed_len,
			sizeof(print_buf) - printed_len, fmt, args);
	va_end(args);

	if TEST_ACTION_STATE(SUSPEND_LOGALL) {
		/* If we didn't print anything, don't do the \n anyway! */
		if (!printed_len)
			return;
		if (restartline)
			printk("\n");
		printk(print_buf);
	} else
		cond_console_print(print_buf, printed_len);
#endif
}

/* printlog
 *
 * Description: This function is a wrapper around printk. It makes the display
 * 		conditional and adds the log level of the message to the front,
 * 		so we can change the detail displayed at runtime.
 * Arguments:	int mask: The debugging section(s) this message belongs to.
 * 		int level: The level of verbosity of this message.
 * 		const char *fmt, ...: Message to be displayed via printk.
 */
//void printlog(int mask, int level, const char *fmt, ...)
void __printlog(const char *fmt, ...)
{
#if defined(DEFAULT_SUSPEND_CONSOLE)
	va_list args;
	int printed_len = 0;

	/* Don't do this if not printing anything - print[no]log get called 
	 * before we prepare the console at resume time */
	handle_loglevel_change();

	va_start(args, fmt);
	printed_len += vsnprintf(print_buf + printed_len, 
			sizeof(print_buf) - printed_len, fmt, args);
	va_end(args);

	printk(print_buf);
#endif
}
#endif

/* prepare_status
 * Description:	Prepare the 'nice display', drawing the header and version,
 * 		along with the current action and perhaps also resetting the
 * 		progress bar.
 * Arguments:	int printalways: Whether to print the action when debugging
 * 		is on.
 * 		int clearbar: Whether to reset the progress bar.
 * 		const char *fmt, ...: The action to be displayed.
 */
void prepare_status(int printalways, int clearbar, const char *fmt, ...)
{
#if defined(DEFAULT_SUSPEND_CONSOLE)
	unsigned char posn[2];
	va_list args;

	if ((NO_OUTPUT_OR_PAUSING) || in_interrupt())
		return;

	if (fmt) {
		va_start(args, fmt);
		lastheader_message_len = vsnprintf(lastheader, 512, fmt, args);
		va_end(args);
	}

	handle_loglevel_change();

	if (console_loglevel >= SUSPEND_ERROR) {

		if (printalways)
			printk("\n** %s\n", lastheader);
	
		/* If nice display is off, there is no progress bar to clear */
		if ((clearbar) && (splash_is_on(fg_console)))
			fb_splash_set_progress(fg_console, 0, 1);
		return;
	}
	
	/* Print version */
	posn[0] = (unsigned char) (0);
	posn[1] = (unsigned char) (video_num_lines);
	move_cursor_to(posn);
	cond_console_print(SWSUSP_CORE_VERSION, strlen(SWSUSP_CORE_VERSION));

	/* Print header */
	posn[0] = (unsigned char) ((video_num_columns - 29) / 2);
	posn[1] = (unsigned char) ((video_num_lines / 3) -4);
	move_cursor_to(posn);

	if (now_resuming) {
		cond_console_print(console_resume, strlen(console_resume));
	} else
		cond_console_print(console_suspend, strlen(console_suspend));
		
	/* Print action */
	posn[1] = (unsigned char) (video_num_lines / 3);
	posn[0] = (unsigned char) 0;
	move_cursor_to(posn);
	
	/* Clear old message */
	for (barposn = 0; barposn < video_num_columns; barposn++) 
		cond_console_print(" ", 1);

	posn[0] = (unsigned char)
		((video_num_columns - lastheader_message_len) / 2);
	move_cursor_to(posn);
	cond_console_print(lastheader, lastheader_message_len);
	
	if (!splash_is_on(fg_console)) {
		/* Draw left bracket of progress bar. */
		posn[0] = (unsigned char) (video_num_columns / 4);
		posn[1]++;
		move_cursor_to(posn);
		cond_console_print("[", 1);

		/* Draw right bracket of progress bar. */
		posn[0] = (unsigned char) 
			(video_num_columns - (video_num_columns / 4) - 1);
		move_cursor_to(posn);
		cond_console_print("]", 1);

		if (clearbar) {
			/* Position at start of progress */
			posn[0] = (unsigned char) (video_num_columns / 4 + 1);
			move_cursor_to(posn);

			/* Clear bar */
			for (barposn = 0; barposn < barwidth; barposn++)
				cond_console_print(" ", 1);
			move_cursor_to(posn);
		}
	}
	
	acquire_console_sem();
	hide_cursor(fg_console);
	release_console_sem();

	if ((clearbar) && (splash_is_on(fg_console)))
		fb_splash_set_progress(fg_console, 0, 1);

	barposn = 0;

	MDELAY(2000);
#endif
}

/* update_status
 *
 * Description: Update the progress bar and (if on) in-bar message.
 * Arguments:	UL value, maximum: Current progress percentage (value/max).
 * 		const char *fmt, ...: Message to be displayed in the middle
 * 		of the progress bar.
 * 		Note that a NULL message does not mean that any previous
 * 		message is erased! For that, you need prepare_status with
 * 		clearbar on.
 * Returns:	Unsigned long: The next value where status needs to be updated.
 * 		This is to reduce unnecessary calls to update_status.
 */
unsigned long update_status(unsigned long value, unsigned long maximum,
		const char *fmt, ...)
{
	unsigned long next_update = 0;
#if defined(DEFAULT_SUSPEND_CONSOLE)
	int bitshift = generic_fls(maximum) - 16;
	unsigned char posn[2];
	va_list args;
	int message_len = 0;

	handle_loglevel_change();

	if (NO_OUTPUT_OR_PAUSING || (!maximum) || (!barwidth))
		return maximum;

	if (value < 0)
		value = 0;

	if (value > maximum)
		value = maximum;

	/* Try to avoid math problems - we can't do 64 bit math here
	 * (and shouldn't need it - anyone got screen resolution
	 * of 65536 pixels or more?) */
	if (bitshift > 0) {
		unsigned long temp_maximum = maximum >> bitshift;
		unsigned long temp_value = value >> bitshift;
		newbarposn = (int) (temp_value * barwidth / temp_maximum);
	} else
		newbarposn = (int) (value * barwidth / maximum);
	
	if (newbarposn < barposn)
		barposn = 0;

	next_update = ((newbarposn + 1) * maximum / barwidth) + 1;

	if (console_loglevel >= SUSPEND_ERROR) {
		if ((splash_is_on(fg_console)) &&
				(newbarposn != barposn)) {
			fb_splash_set_progress(fg_console, 
					value, maximum);
			barposn = newbarposn;
		}
		return next_update;
	}

	/* Update bar */
	if (splash_is_on(fg_console)) {
		if (newbarposn != barposn)
			fb_splash_set_progress(fg_console, value, maximum);
	} else {
		posn[1] = (unsigned char) ((video_num_lines / 3) + 1);

		/* Clear bar if at start */
		if (!barposn) {
			posn[0] = (unsigned char) (video_num_columns / 4 + 1);
			move_cursor_to(posn);
			for (; barposn < barwidth; barposn++)
				cond_console_print(" ", 1);
			barposn = 0;
		}
		posn[0] = (unsigned char) (video_num_columns / 4 + 1 + barposn);
		move_cursor_to(posn);

		for (; barposn < newbarposn; barposn++)
			cond_console_print("-", 1);
	}

	/* Print string in progress bar on loglevel 1 */
	if ((fmt) && (console_loglevel)) {
		va_start(args, fmt);
		message_len = vsnprintf(print_buf, sizeof(print_buf), " ", NULL);
		message_len += vsnprintf(print_buf + message_len,
				sizeof(print_buf) - message_len, fmt, args);
		message_len += vsnprintf(print_buf + message_len,
				sizeof(print_buf) - message_len, " ", NULL);
		va_end(args);

		if (message_len) {
			posn[0] = (unsigned char)
				((video_num_columns - message_len) / 2);
			posn[1] = (unsigned char)
				((video_num_lines / 3) + 1);
			move_cursor_to(posn);
			cond_console_print(print_buf, message_len);
		}
	}

	barposn = newbarposn;
	acquire_console_sem();
	hide_cursor(fg_console);
	release_console_sem();
	
#else
	next_update = maximum;
#endif
	return next_update;
}

static struct waiting_message
{
	int message;
	struct waiting_message * next;
} * waiting_messages = NULL;

/* display_suspend_message
 *
 * Description:	Display a message as a result of the user pressing a key
 * 		and the key being processed in an interrupt handler.
 */

#ifdef CONFIG_SOFTWARE_SUSPEND_VARIATION_ANALYSIS
extern void suspend_calculate_checksums(void);
extern void suspend_check_checksums(void);
#endif

int display_suspend_messages(void)
{
	int did_work = ((waiting_messages != NULL) || (console_loglevel != lastloglevel));
	
	if (console_loglevel != lastloglevel)
		handle_loglevel_change();

	while (waiting_messages) {
		struct waiting_message * this_message = waiting_messages;

		switch(waiting_messages->message) {
			case 1:
				prepare_status(1, 0, "Pausing %s.", 
					TEST_ACTION_STATE(SUSPEND_PAUSE) ?
					"enabled" : "disabled");
				break;
			case 2:
				prepare_status(1, 0, "Rebooting %s.", 
					TEST_ACTION_STATE(SUSPEND_REBOOT) ?
					"enabled" : "disabled");
				break;
			case 3:
				prepare_status(1, 0, "Single step %s.", 
					TEST_ACTION_STATE(SUSPEND_SINGLESTEP) ?
					"enabled" : "disabled");
				break;
			case 4:
				prepare_status(1, 0, "Logging all output %s.",
					TEST_ACTION_STATE(SUSPEND_LOGALL) ?
					"enabled" : "disabled");
				break;
			case 5:
				prepare_status(1, 1,
					"--- ESCAPE PRESSED AGAIN :"
					" TRYING HARDER TO ABORT ---");
				break;
			case 6:
				prepare_status(1, 1, "--- ESCAPE PRESSED :"
						" ABORTING PROCESS ---");
				break;
			case 7:
				prepare_status(1, 0, "Slowing down %s.",
					TEST_ACTION_STATE(SUSPEND_SLOW) ?
					"enabled" : "disabled");
				break;
			case 8:
				prepare_status(1, 0, "Log level changed to %d.",
					console_loglevel);
				break;
#ifdef CONFIG_SOFTWARE_SUSPEND_VARIATION_ANALYSIS
			case 10:
				/* Grab image checksums */
				if (CHECKMASK(SUSPEND_INTEGRITY))
					suspend_calculate_checksums();
				break;
			case 11:
				/* Calculate differences for pageset 1 only */
				if (CHECKMASK(SUSPEND_INTEGRITY))
					suspend_check_checksums();
				break;
#endif
			case 20:
				prepare_status(1, 0, "General messages %s.",
					TEST_DEBUG_STATE(SUSPEND_ANY_SECTION) ?
					"enabled" : "disabled");
				break;
			case 21:
				prepare_status(1, 0, "Freezer messages %s.",
					TEST_DEBUG_STATE(SUSPEND_FREEZER) ?
					"enabled" : "disabled");
				break;
			case 22:
				prepare_status(1, 0, "Eat memory messages %s.",
					TEST_DEBUG_STATE(SUSPEND_EAT_MEMORY) ?
					"enabled" : "disabled");
				break;
			case 23:
				prepare_status(1, 0, "Pageset messages %s.",
					TEST_DEBUG_STATE(SUSPEND_PAGESETS) ?
					"enabled" : "disabled");
				break;
			case 24:
				prepare_status(1, 0, "IO messages %s.",
					TEST_DEBUG_STATE(SUSPEND_IO) ?
					"enabled" : "disabled");
				break;
			case 25:
				prepare_status(1, 0, "Bmap messages %s.",
					TEST_DEBUG_STATE(SUSPEND_BMAP) ?
					"enabled" : "disabled");
				break;
			case 26:
				prepare_status(1, 0, "Swap messages %s.",
					TEST_DEBUG_STATE(SUSPEND_SWAP) ?
					"enabled" : "disabled");
				break;
			case 27:
				prepare_status(1, 0, "Memory messages %s.",
					TEST_DEBUG_STATE(SUSPEND_MEMORY) ?
					"enabled" : "disabled");
				break;
			case 28:
				prepare_status(1, 0, "Range messages %s.",
					TEST_DEBUG_STATE(SUSPEND_RANGES) ?
					"enabled" : "disabled");
				break;
			case 29:
				prepare_status(1, 0, "Memory pool messages %s.",
					TEST_DEBUG_STATE(SUSPEND_MEM_POOL) ?
					"enabled" : "disabled");
				break;
			case 30:
				prepare_status(1, 0, "Nosave messages %s.",
					TEST_DEBUG_STATE(SUSPEND_NOSAVE) ?
					"enabled" : "disabled");
				break;
			case 31:
				prepare_status(1, 0, "Integrity messages %s.",
					TEST_DEBUG_STATE(SUSPEND_INTEGRITY) ?
					"enabled" : "disabled");
				break;
		}

		waiting_messages = this_message->next;
		kfree(this_message);
	}
	return did_work;
}

/* schedule_suspend_message
 *
 * Description:
 * 
 */

void schedule_suspend_message(int message_number)
{
	struct waiting_message * new_message =
		kmalloc(sizeof(struct waiting_message), GFP_ATOMIC);

	if (!new_message) {
		printk("Argh. Unable to allocate memory for "
				"scheduling the display of a message.\n");
		return;
	}

	new_message->message = message_number;
	new_message->next = waiting_messages;

	waiting_messages = new_message;
	return;
}

/* wakeup_suspend
 *
 * Description: Wake up suspend when a key is pressed on the console
 * 		(serial or local) indicating we should continue.
 */

void wakeup_suspend(void)
{
	wake_up_interruptible(&suspend_wait_for_key);
}

/* request_abort_suspend
 *
 * Description:	Handle the user requesting the cancellation of a suspend by
 * 		pressing escape. Note that on a second press, we try a little
 * 		harder, attempting to forcefully thaw processes. This shouldn't
 * 		been needed, and may result in an oops (if we've overwritten
 * 		memory), but has been useful on ocassion.
 * Callers:	Called from drivers/char/keyboard.c or drivers/char/serial.c
 * 		when the user presses escape.
 */
void request_abort_suspend(void)
{
	if ((now_resuming) || (TEST_RESULT_STATE(SUSPEND_ABORT_REQUESTED)))
		return;

	if (TEST_RESULT_STATE(SUSPEND_ABORTED)) {
		schedule_suspend_message(5);
		show_state();
		thaw_processes();
		wakeup_suspend();
	} else {
		schedule_suspend_message(6);
		SET_RESULT_STATE(SUSPEND_ABORTED);
		SET_RESULT_STATE(SUSPEND_ABORT_REQUESTED);
		wakeup_suspend();
	}
}

/* check_shift_keys
 * 
 * Description:	Potentially pause and wait for the user to tell us to continue.
 * 		We normally only pause when @pause is set.
 * Arguments:	int pause: Whether we normally pause.
 * 		char * message: The message to display. Not parameterised
 * 		 because it's normally a constant.
 */

void check_shift_keys(int pause, char * message)
{
#ifdef DEFAULT_SUSPEND_CONSOLE 
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	int displayed_message = 0;
#endif
	
	if (NO_OUTPUT_OR_PAUSING)
		return;

	do {
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
		if (((TEST_ACTION_STATE(SUSPEND_PAUSE) && pause) || 
				(TEST_ACTION_STATE(SUSPEND_SINGLESTEP)))) {
			if (!displayed_message) {
				prepare_status(1, 0, 
				   "%s Press SPACE to continue.%s",
				   message ? message : "",
				   (TEST_ACTION_STATE(SUSPEND_SINGLESTEP)) ? 
				   " Single step on." : "");
				displayed_message = 1;
			}
			interruptible_sleep_on(&suspend_wait_for_key);
		}
#endif
	} while (display_suspend_messages());
#endif
}

extern asmlinkage long sys_ioctl(unsigned int fd, unsigned int cmd, 
		unsigned long arg);

/*
 *
 */

static void suspend_get_dev_console(void)
{
	if (suspend_console_fd > -1)
		return;

	suspend_console_fd = sys_open("/dev/console", O_RDWR | O_NONBLOCK, 0);
	if (suspend_console_fd < 0) {
		sys_mkdir("/dev", 0700);
#ifdef CONFIG_DEVFS_FS
		sys_mount("devfs", "/dev", "devfs", 0, NULL);
		mounted_devfs = 1;
#endif
		suspend_console_fd = sys_open("/dev/console", O_RDWR | O_NONBLOCK, 0);
	}
	if (suspend_console_fd < 0) {
		printk("Can't open /dev/console. Error value was %d.\n",
				suspend_console_fd);
		suspend_console_fd = -1;
		return;
	}

	sys_ioctl(suspend_console_fd, TCGETS, (long)&termios);
	termios.c_lflag &= ~ICANON;
	sys_ioctl(suspend_console_fd, TCSETSF, (long)&termios);
}

/* pm_prepare_console
 *
 * Description:	Prepare a console for use, save current settings.
 * Returns:	Boolean: Whether an error occured. Errors aren't
 * 		treated as fatal, but a warning is printed.
 */
int pm_prepare_console(void)
{
	if (TEST_ACTION_STATE(SUSPEND_NO_OUTPUT))
		/* No output should be produced. */ 
		return 0;

	orig_loglevel = console_loglevel;
	orig_default_message_loglevel = default_message_loglevel;
	if (!now_resuming)
		console_loglevel = suspend_default_console_level;

#ifdef CONFIG_VT_CONSOLE
	orig_fgconsole = fg_console;

	if ((orig_fgconsole != DEFAULT_SUSPEND_CONSOLE) &&
			(!splash_is_on(fg_console))) {
		int wait_result, alloc_result;

		acquire_console_sem();

		alloc_result = vc_allocate(DEFAULT_SUSPEND_CONSOLE);

		release_console_sem();

		/* If we can't have a free VC for now, too bad.*/
		if (alloc_result)
			return 1;

		set_console(DEFAULT_SUSPEND_CONSOLE);

		/* We should have a timeout here */
		wait_result = vt_waitactive(DEFAULT_SUSPEND_CONSOLE);

		if (wait_result) {
			printk("Can't switch virtual consoles.");
			return 1;
		}
	}
	
	suspend_get_dev_console();
	
	clear_display();
	orig_kmsg = kmsg_redirect;
	kmsg_redirect = fg_console;
	prepare_status(0, 0, "");

	default_message_loglevel = 1;
#endif
	if (splash_is_on(fg_console)) {
		if (console_loglevel > 1)
			splash_write_proc("verbose\n", 9);
		else
			splash_write_proc("silent\n", 8);
	}
	return 0;
}

/* suspend_relinquish_console
 *
 * Description: Close our handle on /dev/console. Must be done
 * earlier than pm_restore_console to avoid problems with other
 * processes trying to grab it when thawed.
 */

/* pm_restore_console
 *
 * Description: Restore the settings we saved above.
 */

void pm_restore_console(void)
{
	if (TEST_ACTION_STATE(SUSPEND_NO_OUTPUT))
		return;

#ifdef CONFIG_VT_CONSOLE
	reset_terminal(fg_console, 1);
	termios.c_lflag |= ICANON;
	sys_ioctl(suspend_console_fd, TCSETSF, (long)&termios);
	sys_close(suspend_console_fd);
	suspend_console_fd = -1;

#ifdef CONFIG_DEVFS_FS
	if (mounted_devfs)
		sys_umount("/dev", 0);
#endif

	if (orig_fgconsole != fg_console) {
		set_console(orig_fgconsole);
		/* We should have a timeout here */
		if (vt_waitactive(orig_fgconsole)) {
			printk("Can't switch virtual consoles.");
			return;
		}
	} 

	acquire_console_sem();
	update_screen(fg_console);
	release_console_sem();
	
	kmsg_redirect = orig_kmsg;
#endif

	suspend_default_console_level = console_loglevel;
	console_loglevel = orig_loglevel;
	default_message_loglevel = orig_default_message_loglevel;
	lastloglevel = -1;
	return;
}

#ifdef CONFIG_VT_CONSOLE
/* force_console_redraw
 *
 * Description:	Force a redraw of the console. Necessary after copying the
 * 		original kernel back and when the progress bar is moved
 * 		backwards.
 */
extern int console_blanked;

void force_console_redraw(void)
{

	console_blanked = fg_console + 1;
	//handle_loglevel_change();
	acquire_console_sem();
	unblank_screen();
	update_screen(fg_console);
	release_console_sem();
}

/* post_resume_console_redraw(void)
 *
 * Description:	Redraw the console after copying the original kernel back.
 */

void post_resume_console_redraw(void)
{
	if (splash_is_on(fg_console)) {
		if (console_loglevel < 2)
			splash_write_proc("silent\n", 8);
		else
			splash_write_proc("verbose\n", 9);
	}
}
#endif

/* suspend_early_boot_message()
 * Description:	Complain when the image doesn't appear to match the booted
 * 		kernel. The user may press C to invalidate the image
 * 		and continue booting, or space to reboot. This works from
 * 		either the serial console or normally attached keyboard.
 *
 * 		Note that we come in here from init, while the kernel is
 * 		locked. If we want to get events from the serial console,
 * 		we need to temporarily unlock the kernel.
 *
 * Arguments:	Char *. Pointer to a string explaining why we're moaning.
 */

int suspend_early_boot_message(char *warning_reason)
{
	PRINTPREEMPTCOUNT("At entry to suspend_early_boot_message");
#if defined(CONFIG_VT) || defined(CONFIG_SERIAL_CONSOLE)
#ifdef CONFIG_VT
	kd_mksound(300,HZ/4);
	mdelay(300);
	splash_verbose();
#endif
	printk(KERN_EMERG "=== Software Suspend ===\n\n");
	if (warning_reason) {
		printk(KERN_EMERG "BIG FAT WARNING!! %s\n\n", warning_reason);
		printk(KERN_EMERG "If you want to use the current suspend image, reboot and try\n");
		printk(KERN_EMERG "again with the same kernel that you suspended from. If you want\n");
		printk(KERN_EMERG "to forget that image, continue and the image will be erased.\n");
		printk(KERN_EMERG "Press SPACE to reboot or C to continue booting with this kernel\n");
	} else {
		printk(KERN_EMERG "BIG FAT WARNING!!\n\n");
		printk(KERN_EMERG "You have tried to resume from this image before.\n");
		printk(KERN_EMERG "If it failed once, it will probably fail again.\n");
		printk(KERN_EMERG "Would you like to remove the image and boot normally?\n");
		printk(KERN_EMERG "This will be equivalent to entering noresume2 on the\n");
		printk(KERN_EMERG "kernel command line.\n\n");
		printk(KERN_EMERG "Press SPACE to remove the image or C to continue resuming.\n");
	}
	
	software_suspend_state |= SOFTWARE_SUSPEND_SANITY_CHECK_PROMPT;
	PRINTPREEMPTCOUNT("prior to interruptible sleep on");
	interruptible_sleep_on(&suspend_wait_for_key);
	if ((warning_reason) && (!(software_suspend_state & SOFTWARE_SUSPEND_CONTINUE_REQ)))
		machine_restart(NULL);
	PRINTPREEMPTCOUNT("post interruptible sleep on");
	software_suspend_state &= ~SOFTWARE_SUSPEND_SANITY_CHECK_PROMPT;
#endif // CONFIG_VT or CONFIG_SERIAL_CONSOLE
	return -EPERM;
}

#if defined(CONFIG_SOFTWARE_SUSPEND2) && defined(CONFIG_PROC_FS)
/*
 * User interface specific /proc/suspend entries.
 */

static struct suspend_proc_data proc_params[] = {
	{ .filename			= "default_console_level",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_INTEGER,
	  .data = {
		  .integer = {
			  .variable	= &suspend_default_console_level,
			  .minimum	= 0,
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
			  .maximum	= 7,
#else
			  .maximum	= 1,
#endif

		  }
	  }
	},

	{ .filename			= "enable_escape",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &suspend_action,
			  .bit		= SUSPEND_CAN_CANCEL,
		  }
	  }
	},

	{ .filename			= "no_output",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &suspend_action,
			  .bit		= SUSPEND_NO_OUTPUT,
		  }
	  }
	},
	
#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
	{ .filename			= "debug_sections",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_UL,
	  .data = {
		  .ul = {
			  .variable	= &suspend_debug_state,
			  .minimum	= 0,
			  .maximum	= 2 << 30,
		  }
	  }
	},

	{ .filename			= "log_everything",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &suspend_action,
			  .bit		= SUSPEND_LOGALL,
		  }
	  }
	},
	  
	{ .filename			= "pause_between_steps",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_BIT,
	  .data = {
		  .bit = {
			  .bit_vector	= &suspend_action,
			  .bit		= SUSPEND_PAUSE,
		  }
	  }
	},
#endif

#if defined(CONFIG_FBCON_SPLASHSCREEN) || defined(CONFIG_BOOTSPLASH)
	{ .filename			= "progressbar_granularity_limit",
	  .permissions			= PROC_RW,
	  .type				= SWSUSP_PROC_DATA_INTEGER,
	  .data = {
		  .integer = {
			  .variable	= &bar_granularity_limit,
			  .minimum	= 1,
			  .maximum	= 2000,
		  }
	  }
	}
#endif
};

/* suspend_console_proc_init
 * Description: Boot time initialisation for user interface.
 */
__init void suspend_console_proc_init(void)
{
	int i, numfiles = sizeof(proc_params) / sizeof(struct suspend_proc_data);

	for (i=0; i< numfiles; i++)
		suspend_register_procfile(&proc_params[i]);
}

#endif

#ifdef CONFIG_SOFTWARE_SUSPEND_DEBUG
EXPORT_SYMBOL(__printlog);
#endif
EXPORT_SYMBOL(request_abort_suspend);
EXPORT_SYMBOL(wakeup_suspend);
EXPORT_SYMBOL(prepare_status);
